"""
Main Knowledge Extractor class.

This works by recursively constructing structured prompt-completions where
a pseudo-YAML structure is requested, where the YAML
structure corresponds to a template class.

Describe in the SPIRES manuscript
TODO: add link
"""
import logging
import re
import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import pydantic
import yaml
from linkml_runtime.linkml_model import ClassDefinition, SlotDefinition
from oaklib import BasicOntologyInterface

from ontogpt.engines.knowledge_engine import (
    ANNOTATION_KEY_PROMPT,
    ANNOTATION_KEY_PROMPT_SKIP,
    EXAMPLE,
    FIELD,
    OBJECT,
    KnowledgeEngine,
    chunk_text,
)
from ontogpt.io.yaml_wrapper import dump_minimal_yaml
from ontogpt.templates.core import ExtractionResult

this_path = Path(__file__).parent


RESPONSE_ATOM = Union[str, "ResponseAtom"]  # type: ignore
RESPONSE_DICT = Dict[FIELD, Union[RESPONSE_ATOM, List[RESPONSE_ATOM]]]


@dataclass
class SPIRESEngine(KnowledgeEngine):
    """Knowledge extractor."""

    engine: str = None
    """The engine name."""

    recurse: bool = True
    """If true, then complex non-named entity objects are always recursively parsed.
    If this is false AND the complex object is a pair, then token-based splitting is
    instead used.
    TODO: deprecate this, it's not clear that token-based splitting is better, due to
    the inability to control which tokens GPT will use"""

    sentences_per_window: Optional[int] = None
    """If set, this will split the text into chains of sentences,
    where this determines the maximum number of sentences per chain.
    The results are then merged together."""

    def extract_from_text(
        self,
        text: str,
        cls: ClassDefinition = None,
        object: OBJECT = None,
        show_prompt: bool = False,
    ) -> ExtractionResult:
        """
        Extract annotations from the given text.

        :param text:
        :param cls:
        :param object: optional stub object
        :return:
        """
        if self.sentences_per_window:
            chunks = chunk_text(text, self.sentences_per_window)
            extracted_object = None
            for chunk in chunks:
                raw_text = self._raw_extract(chunk, cls=cls, object=object, show_prompt=show_prompt)
                logging.info(f"RAW TEXT: {raw_text}")
                next_object = self.parse_completion_payload(
                    raw_text, cls, object=object  # type: ignore
                )
                if extracted_object is None:
                    extracted_object = next_object
                else:
                    for k, v in next_object.items():
                        if isinstance(v, list):
                            extracted_object[k] += v
                        else:
                            if k not in extracted_object:
                                extracted_object[k] = v
                            else:
                                extracted_object[k] = v
        else:
            raw_text = self._raw_extract(text=text, cls=cls, object=object, show_prompt=show_prompt)
            logging.info(f"RAW TEXT: {raw_text}")
            extracted_object = self.parse_completion_payload(
                raw_text, cls, object=object  # type: ignore
            )
        return ExtractionResult(
            input_text=text,
            raw_completion_output=raw_text,
            prompt=self.last_prompt,
            extracted_object=extracted_object,
            named_entities=self.named_entities,
        )

    def _extract_from_text_to_dict(self, text: str, cls: ClassDefinition = None) -> RESPONSE_DICT:
        raw_text = self._raw_extract(text=text, cls=cls)
        return self._parse_response_to_dict(raw_text, cls)

    def generate_and_extract(
        self, entity: str, prompt_template: str = "", show_prompt: bool = False, **kwargs
    ) -> ExtractionResult:
        """
        Generate a description using GPT and then extract from it using SPIRES.

        :param entity:
        :param kwargs:
        :return:
        """
        if prompt_template == "":
            prompt_template = "Generate a comprehensive description of {entity}.\n"
        prompt = prompt_template.format(entity=entity)
        if self.client is not None:
            payload = self.client.complete(prompt=prompt, show_prompt=show_prompt)
        else:
            payload = ""
        return self.extract_from_text(payload, **kwargs)

    def iteratively_generate_and_extract(
        self,
        entity: str,
        cache_path: Union[str, Path],
        iteration_slots: List[str],
        adapter: BasicOntologyInterface = None,
        clear=False,
        max_iterations=10,
        prompt_template=None,
        show_prompt: bool = False,
        **kwargs,
    ) -> Iterator[ExtractionResult]:
        def _remove_parenthetical_context(s: str):
            return re.sub(r"\(.*\)", "", s)

        iteration = 0
        if isinstance(cache_path, str):
            cache_path = Path(cache_path)
        if cache_path:
            if cache_path.exists() and not clear:
                db = yaml.safe_load(cache_path.open())
                if "entities_in_queue" not in db:
                    db["entities_in_queue"] = []
            else:
                db = {"processed_entities": [], "entities_in_queue": [], "results": []}
        if entity not in db["processed_entities"]:
            db["entities_in_queue"].append(entity)
        if prompt_template is None:
            prompt_template = (
                "Generate a comprehensive description of {entity}. "
                + "The description should include the information on"
                + " and ".join(iteration_slots)
                + ".\n"
            )
        while db["entities_in_queue"] and iteration < max_iterations:
            iteration += 1
            next_entity = db["entities_in_queue"].pop(0)
            logging.info(f"ITERATION {iteration}, entity={next_entity}")
            # check if entity matches a curie pattern using re
            if re.match(r"^[A-Z]+:[A-Z0-9]+$", next_entity):
                curie = next_entity
                next_entity = adapter.label(next_entity)
            else:
                curie = None
            result = self.generate_and_extract(
                next_entity, prompt_template=prompt_template, show_prompt=show_prompt, **kwargs
            )
            if curie:
                if result.extracted_object:
                    result.extracted_object.id = curie
            db["results"].append(result)
            db["processed_entities"].append(next_entity)
            yield result
            for s in iteration_slots:
                # if s not in result.extracted_object:
                #    raise ValueError(f"Slot {s} not found in {result.extracted_object}")
                vals = getattr(result.extracted_object, s, [])
                if not vals:
                    logging.info("dead-end: no values found for slot")
                    continue
                if not isinstance(vals, list):
                    vals = [vals]
                for val in vals:
                    entity = val
                    if result.named_entities is not None:
                        for ne in result.named_entities:
                            if ne.id == val:
                                entity = ne.label
                                if ne.id.startswith("AUTO"):
                                    # Sometimes the value of some slots will lack
                                    context = next_entity
                                    context = re.sub(r"\(.*\)", "", context)
                                    entity = f"{entity} ({context})"
                                else:
                                    entity = ne.id
                                break
                    queue_deparenthesized = [
                        _remove_parenthetical_context(e) for e in db["entities_in_queue"]
                    ]
                    if (
                        entity not in db["processed_entities"]
                        and entity not in db["entities_in_queue"]
                        and _remove_parenthetical_context(entity) not in queue_deparenthesized
                    ):
                        db["entities_in_queue"].append(entity)
            with open(cache_path, "w") as f:
                # TODO: consider a more robust backend e.g. mongo
                f.write(dump_minimal_yaml(db))

    def generalize(
        self,
        object: Union[pydantic.BaseModel, dict],
        examples: List[EXAMPLE],
        show_prompt: bool = False,
    ) -> ExtractionResult:
        """
        Generalize the given examples.

        :param object:
        :param examples:
        :return:
        """
        cls = self.template_class
        sv = self.schemaview
        prompt = "example:\n"
        for example in examples:
            prompt += f"{self.serialize_object(example)}\n\n"
        prompt += "\n\n===\n\n"
        if isinstance(object, pydantic.BaseModel):
            object = object.dict()
        for k, v in object.items():
            if v:
                slot = sv.induced_slot(k, cls.name)
                prompt += f"{k}: {self._serialize_value(v, slot)}\n"
        logging.debug(f"PROMPT: {prompt}")
        payload = self.client.complete(prompt, show_prompt)
        prediction = self.parse_completion_payload(payload, object=object)
        return ExtractionResult(
            input_text=prompt,
            raw_completion_output=payload,
            # prompt=self.last_prompt,
            results=[prediction],
            named_entities=self.named_entities,
        )

    def map_terms(
        self, terms: List[str], ontology: str, show_prompt: bool = False
    ) -> Dict[str, str]:
        """
        Map the given terms to the given ontology.

        EXPERIMENTAL

        currently GPT-3 does not do so well with this task.

        :param terms:
        :param ontology:
        :return:
        """
        # TODO: make a separate config
        examples = {
            "go": {
                "nucleui": "nucleus",
                "mitochondrial": "mitochondrion",
                "signaling": "signaling pathway",
                "cysteine biosynthesis": "cysteine biosynthetic process",
                "alcohol dehydrogenase": "alcohol dehydrogenase activity",
            },
            "uberon": {
                "feet": "pes",
                "forelimb, left": "left forelimb",
                "hippocampus": "Ammons horn",
            },
        }
        ontology = ontology.lower()
        if ontology in examples:
            example = examples[ontology]
        else:
            example = examples["uberon"]
        prompt = "Normalize the following semicolon separated\
            list of terms to the {ontology.upper()} ontology\n\n"
        prompt += "For example:\n\n"
        for k, v in example.items():
            prompt += f"{k}: {v}\n"
        prompt += "===\n\nTerms:"
        prompt += "; ".join(terms)
        prompt += "===\n\n"
        payload = self.client.complete(prompt, show_prompt)
        # outer parse
        best_results: List[str] = []
        for sep in ["\n", "; "]:
            results = payload.split(sep)
            if len(results) > len(best_results):
                best_results = results

        def normalize(s: str) -> str:
            s = s.strip()
            s.replace("_", " ")
            return s.lower()

        mappings = {}
        for result in best_results:
            if ":" not in result:
                logging.error(f"Count not parse result: {result}")
                continue
            k, v = result.strip().split(":", 1)
            k = k.strip()
            v = v.strip()
            for t in terms:
                if normalize(t) == normalize(k):
                    mappings[t] = v
                    break
        for t in terms:
            if t not in mappings:
                logging.warning(f"Could not map term: {t}")
        return mappings

    def serialize_object(self, example: EXAMPLE, cls: ClassDefinition = None) -> str:
        if cls is None:
            cls = self.template_class
        if isinstance(example, str):
            return example
        if isinstance(example, pydantic.BaseModel):
            example = example.dict()
        lines = []
        sv = self.schemaview
        for k, v in example.items():
            if not v:
                continue
            slot = sv.induced_slot(k, cls.name)
            v_serialized = self._serialize_value(v, slot)
            lines.append(f"{k}: {v_serialized}")
        return "\n".join(lines)

    def _serialize_value(self, val: Any, slot: SlotDefinition) -> str:
        if val is None:
            return ""
        if isinstance(val, list):
            return "; ".join([self._serialize_value(v, slot) for v in val if v])
        if isinstance(val, dict):
            return " - ".join([self._serialize_value(v, slot) for v in val.values() if v])
        sv = self.schemaview
        if slot.range in sv.all_classes():
            if self.labelers:
                labelers = list(self.labelers)
            else:
                labelers = []
            labelers += self.get_annotators(sv.get_class(slot.range))
            if labelers:
                for labeler in labelers:
                    label = labeler.label(val)
                    if label:
                        return label
        return val

    def _raw_extract(
        self,
        text,
        cls: ClassDefinition = None,
        object: OBJECT = None,
        show_prompt: bool = False,
    ) -> str:
        """
        Extract annotations from the given text.

        :param text:
        :return:
        """
        prompt = self.get_completion_prompt(cls=cls, text=text, object=object)
        self.last_prompt = prompt
        payload = self.client.complete(prompt=prompt, show_prompt=show_prompt)
        return payload

    def get_completion_prompt(
        self, cls: ClassDefinition = None, text: str = "", object: OBJECT = None
    ) -> str:
        """Get the prompt for the given template."""
        if cls is None:
            cls = self.template_class
        if not text or ("\n" in text or len(text) > 60):
            prompt = (
                "From the text below, extract the following entities in the following format:\n\n"
            )
        else:
            prompt = "Split the following piece of text into fields in the following format:\n\n"
        for slot in self.schemaview.class_induced_slots(cls.name):
            if ANNOTATION_KEY_PROMPT_SKIP in slot.annotations:
                continue
            if ANNOTATION_KEY_PROMPT in slot.annotations:
                slot_prompt = slot.annotations[ANNOTATION_KEY_PROMPT].value
            elif slot.description:
                slot_prompt = slot.description
            else:
                if slot.multivalued:
                    slot_prompt = f"semicolon-separated list of {slot.name}s"
                else:
                    slot_prompt = f"the value for {slot.name}"
            if slot.range in self.schemaview.all_enums():
                enum_def = self.schemaview.get_enum(slot.range)
                pvs = [str(k) for k in enum_def.permissible_values.keys()]
                slot_prompt += f"Must be one of: {', '.join(pvs)}"
            prompt += f"{slot.name}: <{slot_prompt}>\n"
        # prompt += "Do not answer if you don't know\n\n"
        prompt = f"{prompt}\n\nText:\n{text}\n\n===\n\n"
        if object:
            if cls is None:
                cls = self.template_class
            if isinstance(object, pydantic.BaseModel):
                object = object.dict()
            for k, v in object.items():
                if v:
                    slot = self.schemaview.induced_slot(k, cls.name)
                    prompt += f"{k}: {self._serialize_value(v, slot)}\n"
        return prompt

    def _parse_response_to_dict(
        self, results: str, cls: ClassDefinition = None
    ) -> Optional[RESPONSE_DICT]:
        """
        Parse the pseudo-YAML response from OpenAI into a dictionary object.

        E.g.

            foo: a; b; c

        becomes

            {"foo": ["a", "b", "c"]}

        :param results:
        :return:
        """
        lines = results.splitlines()
        ann = {}
        promptable_slots = self.promptable_slots(cls)
        for line in lines:
            line = line.strip()
            if not line:
                continue
            if ":" not in line:
                if len(promptable_slots) == 1:
                    slot = promptable_slots[0]
                    logging.warning(
                        f"Coercing to YAML-like with key {slot.name}: Original line: {line}"
                    )
                    line = f"{slot.name}: {line}"
                else:
                    logging.error(f"Line '{line}' does not contain a colon; ignoring")
                    return None
            r = self._parse_line_to_dict(line, cls)
            if r is not None:
                field, val = r
                ann[field] = val
        return ann

    def _parse_line_to_dict(
        self, line: str, cls: ClassDefinition = None
    ) -> Optional[Tuple[FIELD, RESPONSE_ATOM]]:
        if cls is None:
            cls = self.template_class
        sv = self.schemaview
        # each line is a key-value pair
        logging.info(f"PARSING LINE: {line}")
        field, val = line.split(":", 1)
        # Field nornalization:
        # The LLML may mutate the output format somewhat,
        # randomly pluralizing or replacing spaces with underscores
        field = field.lower().replace(" ", "_")
        logging.debug(f"  FIELD: {field}")
        cls_slots = sv.class_slots(cls.name)
        slot = None
        if field in cls_slots:
            slot = sv.induced_slot(field, cls.name)
        else:
            # TODO: check this
            if field.endswith("s"):
                field = field[:-1]
            if field in cls_slots:
                slot = sv.induced_slot(field, cls.name)
        if not slot:
            logging.error(f"Cannot find slot for {field} in {line}")
            # raise ValueError(f"Cannot find slot for {field} in {line}")
            return None
        if not val:
            msg = f"Empty value in key-value line: {line}"
            if slot.required:
                raise ValueError(msg)
            if slot.recommended:
                logging.warning(msg)
            return None
        inlined = slot.inlined
        slot_range = sv.get_class(slot.range)
        if not inlined:
            if slot.range in sv.all_classes():
                inlined = sv.get_identifier_slot(slot_range.name) is None
        val = val.strip()
        if slot.multivalued:
            vals = [v.strip() for v in val.split(";")]
        else:
            vals = [val]
        vals = [val for val in vals if val]
        logging.debug(f"SLOT: {slot.name} INL: {inlined} VALS: {vals}")
        if inlined:
            transformed = False
            slots_of_range = sv.class_slots(slot_range.name)
            if self.recurse or len(slots_of_range) > 2:
                logging.debug(f"  RECURSING ON SLOT: {slot.name}, range={slot_range.name}")
                vals = [
                    self._extract_from_text_to_dict(v, slot_range) for v in vals  # type: ignore
                ]
            else:
                for sep in [" - ", ":", "/", "*", "-"]:
                    if all([sep in v for v in vals]):
                        vals = [
                            dict(zip(slots_of_range, v.split(sep, 1))) for v in vals  # type: ignore
                        ]
                        for v in vals:
                            for k in v.keys():  # type: ignore
                                v[k] = v[k].strip()  # type: ignore
                        transformed = True
                        break
                if not transformed:
                    logging.warning(f"Did not find separator in {vals} for line {line}")
                    return None
        # transform back from list to single value if not multivalued
        if slot.multivalued:
            final_val = vals
        else:
            if len(vals) != 1:
                logging.error(f"Expected 1 value for {slot.name} in '{line}' but got {vals}")
            final_val = vals[0]  # type: ignore
        return field, final_val

    def parse_completion_payload(
        self, results: str, cls: ClassDefinition = None, object: dict = None
    ) -> pydantic.BaseModel:
        """
        Parse the completion payload into a pydantic class.

        :param results:
        :param cls:
        :param object: stub object
        :return:
        """
        raw = self._parse_response_to_dict(results, cls)
        logging.debug(f"RAW: {raw}")
        if object:
            raw = {**object, **raw}
        self._auto_add_ids(raw, cls)
        return self.ground_annotation_object(raw, cls)

    def _auto_add_ids(self, ann: RESPONSE_DICT, cls: ClassDefinition = None) -> None:
        if ann is None:
            return
        if cls is None:
            cls = self.template_class
        for slot in self.schemaview.class_induced_slots(cls.name):
            if slot.identifier:
                if slot.name not in ann:
                    auto_id = str(uuid.uuid4())
                    auto_prefix = self.auto_prefix
                    if slot.range == "uriorcurie" or slot.range == "uri":
                        ann[slot.name] = f"{auto_prefix}:{auto_id}"
                    else:
                        ann[slot.name] = auto_id

    def ground_annotation_object(
        self, ann: RESPONSE_DICT, cls: ClassDefinition = None
    ) -> Optional[pydantic.BaseModel]:
        """Ground the direct parse of the OpenAI payload.

        The raw openAI payload is a YAML-like string, which is parsed to
        a response dictionary.

        This dictionary is then grounded, using this method

        :param ann: Raw annotation object
        :param cls: schema class the ground object should instantiate
        :return: Grounded annotation object
        """
        logging.debug(f"Grounding annotation object {ann}")
        if cls is None:
            cls = self.template_class
        sv = self.schemaview
        new_ann: Dict[str, Any] = {}
        if ann is None:
            logging.error(f"Cannot ground None annotation, cls={cls.name}")
            return None
        for field, vals in ann.items():
            if isinstance(vals, list):
                multivalued = True
            else:
                multivalued = False
                vals = [vals]
            slot = sv.induced_slot(field, cls.name)
            rng_cls = sv.get_class(slot.range)
            enum_def = None
            if slot.range:
                if slot.range in self.schemaview.all_enums():
                    enum_def = self.schemaview.get_enum(slot.range)
            new_ann[field] = []
            logging.debug(f"FIELD: {field} SLOT: {slot.name}")
            for val in vals:
                if not val:
                    continue
                logging.debug(f"   VAL: {val}")
                if isinstance(val, tuple):
                    # special case for pairs
                    sub_slots = sv.class_induced_slots(rng_cls.name)
                    obj = {}
                    for i in range(0, len(val)):
                        sub_slot = sub_slots[i]
                        sub_rng = sv.get_class(sub_slot.range)
                        if not sub_rng:
                            logging.error(f"Cannot find range for {sub_slot.name}")
                        result = self.normalize_named_entity(val[i], sub_slot.range)
                        obj[sub_slot.name] = result
                elif isinstance(val, dict):
                    # recurse
                    obj = self.ground_annotation_object(val, rng_cls)
                else:
                    obj = self.normalize_named_entity(val, slot.range)  # type: ignore
                if enum_def:
                    found = False
                    logging.info(f"Looking for {obj} in {enum_def.name}")
                    for k, _pv in enum_def.permissible_values.items():
                        if type(obj) is str and type(k) is str:
                            if obj.lower() == k.lower():  # type: ignore
                                obj = k  # type: ignore
                                found = True
                                break
                    if not found:
                        logging.info(f"Cannot find enum value for {obj} in {enum_def.name}")
                        obj = None
                if multivalued:
                    new_ann[field].append(obj)
                else:
                    new_ann[field] = obj
        logging.debug(f"Creating object from dict {new_ann}")
        logging.info(new_ann)
        py_cls = self.template_module.__dict__[cls.name]
        return py_cls(**new_ann)



"""Streamlist web app for spindoctor."""
# Import necessary libraries
import re

import streamlit as st
from oaklib import get_adapter

from ontogpt.engines import create_engine
from ontogpt.engines.enrichment import EnrichmentEngine, GeneDescriptionSource
from ontogpt.utils.gene_set_utils import GeneSet

MODEL_GPT_3_5_TURBO = "gpt-3.5-turbo"
MODEL_GPT_3_5_TURBO_INSTRUCT = "gpt-3.5-turbo-instruct"
MODEL_GPT_4 = "gpt-4"

go = get_adapter("sqlite:obo:go")

# Title of the app
st.title("SPINDOCTOR")
st.caption("A tool for summarizing gene sets using GPT")

col1, col2 = st.columns(2)

# Text area for name input
gene_symbols = col1.text_area("Enter a list of human gene symbols")

model = col1.selectbox(
    "Select the model:", (MODEL_GPT_3_5_TURBO, MODEL_GPT_3_5_TURBO_INSTRUCT, MODEL_GPT_4)
)

source = col1.selectbox(
    "Select the gene description source:",
    (
        GeneDescriptionSource.ONTOLOGICAL_SYNOPSIS.value,
        GeneDescriptionSource.NARRATIVE_SYNOPSIS.value,
        GeneDescriptionSource.NONE.value,
    ),
)

openai_api_key = col1.text_input(
    "OpenAI API Key:",
    placeholder="(sk-...) Press [Enter] to submit.",
)

# Button for parsing and displaying the names
if col1.button("Summarize genes"):
    gene_symbols = [symbol.strip() for symbol in re.split(r"[\-,;\s]+", gene_symbols)]
    gene_set = GeneSet(name="TEMP", gene_symbols=gene_symbols)
    ke = create_engine(None, EnrichmentEngine, model=model)
    if openai_api_key:
        ke.set_api_key(openai_api_key)
    if not isinstance(ke, EnrichmentEngine):
        raise ValueError(f"Expected EnrichmentEngine, got {type(ke)}")
    source_pv = GeneDescriptionSource(source)
    col1.write("Analyzing, please wait...")
    results = ke.summarize(gene_set, gene_description_source=source_pv)
    col1.header("Genes")
    for gene_id in gene_set.gene_ids:
        col1.write(f" * {gene_id}")
    # st.write("## Term Strings")
    # for term_string in results.term_strings:
    #    st.write(f" * {term_string}")
    col2.write("## Terms")
    for term_id in results.term_ids:
        if term_id.startswith("GO:"):
            lbl = go.label(term_id)
            col2.markdown(f" * [{term_id}](https://bioregistry.io/{term_id}) - _{lbl}_")
        else:
            col2.markdown(f" * UNPARSED {term_id}")
    col2.header("Summary")
    col2.caption(results.summary)



"""Pubmed Client."""
import logging
import time
from dataclasses import dataclass
from typing import Generator, List, Tuple, Union
from urllib import parse

import inflection
import requests
from bs4 import BeautifulSoup
from oaklib.utilities.apikey_manager import get_apikey_value

PMID = str
TITLE_WEIGHT = 5
MAX_PMIDS = 50
RETRY_MAX = 2

PUBMED = "pubmed"
EUTILS_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/"


def _normalize(s: str) -> str:
    return inflection.singularize(s).lower()


def _score_paper(paper: str, keywords: List[str]) -> Tuple[PMID, int]:
    """Assign a quality score to a PubMed entry.

    Input needs to be XML so it can be parsed by component.
    :param paper: string, the text of the paper, as XML
    :param keywords: list of keywords to use in scoring
    :return: tuple of article ID (PMID) and score
    """
    # Parse the paper by component first
    soup = BeautifulSoup(paper, "xml")
    for pa in soup.find_all(["PubmedArticle", "PubmedBookArticle"]):  # This should be one exactly
        ti = pa.find("ArticleTitle").text
        pmid = pa.find("ArticleId", IdType="pubmed").text
        if pa.find("Abstract"):  # Document may not have abstract
            ab = pa.find("Abstract").text
        else:
            ab = ""

    title_score = _score_text(ti, keywords)
    abstract_score = _score_text(ab, keywords)
    logging.info(f"Scored {pmid} {ti} with TS={title_score} AS={abstract_score} ")
    score = title_score * TITLE_WEIGHT + abstract_score
    return (pmid, score)


def _score_text(text: str, keywords: List[str]) -> int:
    """Assign a quality score to a text entry.

    A text entry not mentioning any of its keywords has a score of zero.
    Score increases by one for each keyword mentioned in the text.
    :param text: The text to assign the score to
    :param keywords: The keywords to use in scoring
    """
    text = text.lower()
    if not text:
        return -100
    score = 0
    for kw in keywords:
        if kw in text:
            score += 1
    return score


def clean_pmids(ids: list[PMID]) -> list[PMID]:
    """Remove prefixes from a list of PMIDs, returning the new list."""
    clean_ids = [id.replace("PMID:", "", 1) for id in ids]
    return clean_ids


@dataclass
class PubmedClient:
    """A client for the Pubmed API.

    This class is a wrapper around the Entrez API.
    """

    # TODO: this doesn't need to be hardcoded
    # and may vary based on the model in use
    max_text_length = 3000

    try:
        email = get_apikey_value("ncbi-email")
    except ValueError:
        email = None
        logging.info("Email for NCBI API not found.")

    try:
        ncbi_key = get_apikey_value("ncbi-key")
    except ValueError:
        ncbi_key = None
        logging.info("NCBI API key not found. Will use no key.")

    def get_pmids(self, term: str) -> List[str]:
        """Search PubMed and retrieve a list of PMIDs matching the search term.

        :param term: The search term to query PubMed.
        :return: A list of PMIDs matching the search term.
        """
        pmids = []

        batch_size = 5000

        search_url = EUTILS_URL + "esearch.fcgi"

        logging.info(f"Finding count of PMIDs matching search term {term}...")
        # If retmax==0, we get only the size of the search result in count of PMIDs
        if self.email and self.ncbi_key:
            params = {
                "db": PUBMED,
                "term": term,
                "retmode": "json",
                "retmax": 0,
                "email": self.email,
                "api_key": self.ncbi_key,
            }
        else:
            params = {"db": PUBMED, "term": term, "retmode": "json", "retmax": 0}

        # We are explicit with the delimiters in this query,
        # no percent encoding allowed. This mostly just makes it more human readable
        response = requests.get(search_url, params=parse.urlencode(params, safe=","))

        if response.status_code == 200:
            data = response.json()
            resultcount = int(data["esearchresult"]["count"])
            logging.info(f"Search returned {resultcount} PMIDs matching search term {term}")
        elif response.status_code == 429:
            logging.error("Too many requests to NCBI API. Try again later, or use API key.")
        else:
            logging.error("Encountered error in searching PubMed:", response.status_code)

        if resultcount > 9999:
            logging.warning("PubMed limits search results to 9999 records.")
        resultcount = 9999

        # Now we get the list of PMIDs, iterating as needed
        logging.info(f"Retrieving PMIDs matching search term {term}...")
        for retstart in range(0, resultcount, batch_size):
            params["retstart"] = retstart
            params["retmax"] = batch_size

            response = requests.get(search_url, params=parse.urlencode(params, safe=","))

            trying = True
            try_count = 0
            while trying:
                if response.status_code == 200:
                    data = response.json(strict=False)
                    try:
                        these_ids = data["esearchresult"]["idlist"]
                        pmids.extend(these_ids)
                    except KeyError:  # Likely an error message.
                        errortext = data["esearchresult"]["ERROR"]
                        logging.error(f"Response: {errortext}")
                    trying = False
                else:
                    logging.error(f"Encountered error in searching PubMed: {response.status_code}")
                    try_count = try_count + 1
                    if try_count < RETRY_MAX:
                        logging.info("Trying again...")
                        time.sleep(1)
                    else:
                        logging.info(f"Giving up - last status code {response.status_code}")
                        trying = False
        logging.info("Retrieved all PMIDs.")

        return pmids

    def text(
        self, ids: Union[list[PMID], PMID], raw=False, autoformat=True, pubmedcental=False
    ) -> Union[list[str], str]:
        """Get the text of one or more papers from their PMIDs.

        :param ids: List of PubMed IDs, or string with single PMID
        :param raw: if True, do not parse the xml, just return the raw output with tags
        :param autoformat: if True include title and abstract concatenated
        :param pubmedcentral: if True, retreive text from PubMed Central where possible
        :return: the text of a single entry, or a list of strings for text of multiple entries
        """
        batch_size = 200

        # Check if the PMID(s) can be parsed
        # and remove prefix if present
        if isinstance(ids, PMID):  # If it's a single PMID
            ids = [ids]
            singledoc = True
        else:
            singledoc = False
        clean_ids = clean_pmids(ids)
        ids = clean_ids

        # this will store the document data
        xml_data = ""

        # Check if we have enough IDs to require epost
        # This is stated as being 10000
        # But in practice, the API won't even accept more than ~2500
        # chars worth of IDs, or about 280 to 320 PMIDs.
        # So instead, we iterate.
        use_post = False
        if len(ids) > 275:
            use_post = True

        fetch_url = EUTILS_URL + "efetch.fcgi"

        if not use_post:
            if self.email and self.ncbi_key:
                params = {
                    "db": PUBMED,
                    "id": ",".join(ids),
                    "rettype": "xml",
                    "retmode": "xml",
                    "email": self.email,
                    "api_key": self.ncbi_key,
                }
            else:
                params = {"db": PUBMED, "id": ",".join(ids), "rettype": "xml", "retmode": "xml"}

            for retstart in range(0, len(ids), batch_size):
                params["retstart"] = retstart
                params["retmax"] = batch_size

                response = requests.get(fetch_url, params=parse.urlencode(params, safe=","))

                trying = True
                try_count = 0
                while trying:
                    if response.status_code == 200:
                        xml_data = xml_data + "\n" + response.text
                        trying = False
                    else:
                        logging.error(
                            f"Encountered error in fetching from PubMed: {response.status_code}"
                        )
                        try_count = try_count + 1
                        if try_count < RETRY_MAX:
                            logging.info("Trying again...")
                            time.sleep(1)
                        else:
                            logging.info(f"Giving up - last status code {response.status_code}")
                            trying = False
            logging.info("Retrieved document data.")

        else:
            # Do post request first, iterating as needed
            post_url = EUTILS_URL + "epost.fcgi"
            if self.email and self.ncbi_key:
                params = {
                    "db": PUBMED,
                    "id": "",
                    "email": self.email,
                    "api_key": self.ncbi_key,
                    "WebEnv": "",
                }
            else:
                params = {"db": PUBMED, "id": ",".join(ids), "WebEnv": ""}

            query_keys = []

            # Need to batch the IDs manually for this endpoint
            id_batches = [ids[pmid : pmid + batch_size] for pmid in range(0, len(ids), batch_size)]

            for batch in id_batches:
                params["id"] = ",".join(batch)

                response = requests.post(post_url, params=parse.urlencode(params, safe=","))

                # Get a webenv the first time, then reuse on subsequent requests
                # import pdb; pdb.set_trace()
                if params["WebEnv"] == "":
                    webenv = response.text.split("<WebEnv>")[1].split("</WebEnv>")[0]
                    params["WebEnv"] = webenv
                querykey = response.text.split("<QueryKey>")[1].split("</QueryKey>")[0]
                query_keys.append(querykey)

                trying = True
                try_count = 0
                while trying:
                    if response.status_code == 200:
                        trying = False
                    else:
                        logging.error(
                            f"Encountered error in posting IDs to PubMed: {response.status_code}"
                        )
                        try_count = try_count + 1
                        if try_count < RETRY_MAX:
                            logging.info("Trying again...")
                            time.sleep(0.5)
                        else:
                            logging.info(f"Giving up - last status code {response.status_code}")
                            trying = False
            logging.info(f"Posted PMIDs to WebEnv:{webenv}.")

            # Now do the fetch
            params = {
                "db": PUBMED,
                "query_key": "",
                "WebEnv": webenv,
                "rettype": "xml",
            }

            # Iterate through the query keys we have
            for this_key in query_keys:
                params["query_key"] = this_key

                response = requests.get(fetch_url, params=parse.urlencode(params, safe=","))

                trying = True
                try_count = 0
                while trying:
                    if response.status_code == 200:
                        xml_data = xml_data + "\n" + response.text
                        trying = False
                    else:
                        logging.error(
                            f"Encountered error in fetching from PubMed: {response.status_code}"
                        )
                        try_count = try_count + 1
                        if try_count < RETRY_MAX:
                            logging.info("Trying again...")
                            time.sleep(0.5)
                        else:
                            logging.info(f"Giving up - last status code {response.status_code}")
                            trying = False
            logging.info("Retrieved document data.")

        # Parse that xml - this returns a list of strings
        # if raw is True, the tags are kept, but we still get a list of docs
        # and we don't truncate them
        these_docs = self.parse_pmxml(
            xml=xml_data, raw=raw, autoformat=autoformat, pubmedcentral=pubmedcental
        )

        txt = []
        onetxt = ""
        for doc in these_docs:
            if len(doc) > self.max_text_length and not raw:
                logging.warning(
                    f'Truncating entry beginning "{doc[:50]}" to {str(self.max_text_length)} chars'
                )
                shortdoc = doc[0 : self.max_text_length]
                txt.append(shortdoc)
            else:
                txt.append(doc)
            if singledoc and not pubmedcental:
                onetxt = txt[0]

        if len(onetxt) > 0:
            return onetxt
        else:
            return txt

    def pmc_text(self, pmc_id: str) -> str:
        """Get the text of one PubMed Central entry.

        Don't parse further here - just get the raw response.
        :param pmc_id: List of PubMed IDs, or string with single PMID
        :return: the text of a single entry as XML
        """
        xml_data = ""

        fetch_url = EUTILS_URL + "efetch.fcgi"

        if self.email and self.ncbi_key:
            params = {
                "db": "pmc",
                "id": pmc_id,
                "rettype": "xml",
                "retmode": "xml",
                "email": self.email,
                "api_key": self.ncbi_key,
            }
        else:
            params = {"db": "pmc", "id": pmc_id, "rettype": "xml", "retmode": "xml"}

        response = requests.get(fetch_url, params=parse.urlencode(params, safe=","))

        trying = True
        try_count = 0
        while trying:
            if response.status_code == 200:
                xml_data = response.text
                trying = False
            else:
                logging.error(
                    f"Encountered error in fetching from PubMed Central: {response.status_code}"
                )
                try_count = try_count + 1
                if try_count < RETRY_MAX:
                    logging.info("Trying again...")
                    time.sleep(1)
                else:
                    logging.info(f"Giving up - last status code {response.status_code}")
                    trying = False
        logging.info(f"Retrieved PubMed Central document data for {pmc_id}.")

        return xml_data

    def search(self, term: str, keywords: List[str]) -> Generator[PMID, None, None]:
        """Get the quality-scored text of PubMed papers relating to a search term and keywords.

        This generator yields PMIDs. Note this uses the MAX_PMIDS value
        to determine how many documents to collect.
        :param term: search term, a string
        :param keywords: keywords, a list of strings
        :return: PMIDs corresponding to the search term and keywords
        """
        if keywords:
            keywords = [_normalize(kw) for kw in keywords]
            term = f"({term}) AND ({' OR '.join(keywords)})"
        else:
            keywords = []

        logging.info(f"Searching for {term}...")

        esr = self.get_pmids(term=term)

        paset = self.text(ids=esr[0:MAX_PMIDS], raw=True)

        scored_papers = [(_score_paper(paper, keywords), paper) for paper in paset]
        scored_papers.sort(key=lambda x: x[1][0], reverse=True)

        for id_and_score, _paper in scored_papers:
            pmid = id_and_score[0]
            score = id_and_score[1]
            logging.debug(f"Yielding {pmid} with score {score} ")
            yield f"{pmid}"

    def parse_pmxml(self, xml: str, raw: bool, autoformat: bool, pubmedcentral: bool) -> List[str]:
        """Extract structured text from PubMed and PubMed Central XML.

        :param xml: One or more xml entries, as string
        :param raw: if True, do not parse the xml beyond separating documents
        :param autoformat: if True include title and abstract concatenated
            Otherwise the output will include ALL text contents besides XML tags
        :param pubmedcentral: if True replace abstract with PubMed Central text
            If there isn't a PMC ID, just use the abstract.
            If there is a PMC ID, use the abstract AND chunk the body text.
            This means the same ID may have multiple entries and may require multiple
            queries to the LLM.
        :return: a list of strings, one per entry
        """
        docs = []

        # Preprocess the string to ensure it's valid xml
        if not raw:
            logging.info("Preprocessing all xml entries...")
            header = "\n".join(xml.split("\n", 3)[0:3])
            pmas_opener = "<PubmedArticleSet>"
            pmas_closer = "</PubmedArticleSet>"
            for remove_string in [header, pmas_opener, pmas_closer]:
                xml = xml.replace(remove_string, "\n")
            xml = pmas_opener + xml + pmas_closer

        soup = BeautifulSoup(xml, "xml")

        logging.info("Parsing all xml entries...")
        for pa in soup.find_all(["PubmedArticle", "PubmedBookArticle"]):
            # First check the PMID, and if requested, any PMC ID
            pmid = ""
            if pa.find("PMID"):  # If this is missing something has gone Wrong
                pmid = pa.find("PMID").text
            pmc_id = ""
            has_pmc_id = False
            try:  # There's a chance that this entry is missing one or more fields below
                if (
                    pa.find("PubmedData").find("ArticleIdList").find("ArticleId", {"IdType": "pmc"})
                    and pubmedcentral
                ):
                    pmc_id = (
                        pa.find("PubmedData")
                        .find("ArticleIdList")
                        .find("ArticleId", {"IdType": "pmc"})
                        .text
                    )
                    has_pmc_id = True
            except AttributeError:
                logging.info(f"PubMed entry {pmid} is missing the expected PubMedData fields.")
            if autoformat and not raw and not has_pmc_id:  # No PMC ID - just use title+abstract
                ti = ""
                if pa.find("ArticleTitle"):
                    ti = pa.find("ArticleTitle").text
                ab = ""
                if pa.find("Abstract"):  # Document may not have abstract
                    ab = pa.find("Abstract").text
                kw = [""]
                if pa.find("KeywordList"):  # Document may not have MeSH terms or keywords
                    kw = [tag.text for tag in pa.find_all("Keyword")]
                txt = f"Title: {ti}\nKeywords: {'; '.join(kw)}\nPMID: {pmid}\nAbstract: {ab}"
                docs.append(txt)
            elif autoformat and not raw and has_pmc_id:  # PMC ID - get and use that text instead
                fulltext = self.pmc_text(pmc_id)
                fullsoup = BeautifulSoup(fulltext, "xml")
                body = ""
                if fullsoup.find("pmc-articleset").find("article").find("body"):
                    body = fullsoup.find("pmc-articleset").find("article").find("body").text
                    body = body.replace("\n", " ")
                ti = ""
                if pa.find("ArticleTitle"):
                    ti = pa.find("ArticleTitle").text
                if pa.find("Abstract"):  # Document may not have abstract
                    body = pa.find("Abstract").text + body
                kw = [""]
                if pa.find("KeywordList"):  # Document may not have MeSH terms or keywords
                    kw = [tag.text for tag in pa.find_all("Keyword")]

                id_txt = f"Title: {ti}\nKeywords: {'; '.join(kw)}\nPMID: {pmid}\nPMCID: {pmc_id}\n"
                full_max_len = self.max_text_length - len(id_txt)
                chunktxt = [body[i : i + full_max_len] for i in range(0, len(body), full_max_len)]
                for txt in chunktxt:
                    docs.append(id_txt + txt)
                    logging.warning(
                        f'Truncating entry containing "{txt[:50]}" to {self.max_text_length} chars'
                    )
            elif raw:
                docs.append(str(pa))
            else:
                docs.append(soup.get_text())

        return docs



from __future__ import annotations
from datetime import datetime, date
from enum import Enum
from typing import List, Dict, Optional, Any, Union
from pydantic import BaseModel as BaseModel, Field
import sys
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


metamodel_version = "None"
version = "None"

class ConfiguredBaseModel(BaseModel,
                validate_assignment = True,
                validate_default = True,
                extra = 'forbid',
                arbitrary_types_allowed = True,
                use_enum_values = True):
    pass


class NullDataOptions(str, Enum):
    
    
    UNSPECIFIED_METHOD_OF_ADMINISTRATION = "UNSPECIFIED_METHOD_OF_ADMINISTRATION"
    
    NOT_APPLICABLE = "NOT_APPLICABLE"
    
    NOT_MENTIONED = "NOT_MENTIONED"
    
    

class Study(ConfiguredBaseModel):
    
    environments: Optional[List[str]] = Field(default_factory=list)
    causal_relationships: Optional[List[CausalRelationship]] = Field(default_factory=list)
    variables: Optional[List[str]] = Field(default_factory=list)
    treatments: Optional[List[str]] = Field(default_factory=list)
    measurements: Optional[List[Measurement]] = Field(default_factory=list)
    location: Optional[List[str]] = Field(default_factory=list, description="""the sites at which the study was conducted""")
    environmental_material: Optional[List[str]] = Field(default_factory=list, description="""the environmental material that was sampled""")
    sequencing_technologies: Optional[str] = Field(None)
    organisms: Optional[List[str]] = Field(default_factory=list, description="""semicolon-separated list of all studied organism taxons""")
    

class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    input_id: Optional[str] = Field(None)
    input_title: Optional[str] = Field(None)
    input_text: Optional[str] = Field(None)
    raw_completion_output: Optional[str] = Field(None)
    prompt: Optional[str] = Field(None)
    extracted_object: Optional[Any] = Field(None, description="""The complex objects extracted from the text""")
    named_entities: Optional[List[Any]] = Field(default_factory=list, description="""Named entities extracted from the text""")
    

class NamedEntity(ConfiguredBaseModel):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Location(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class EnvironmentalMaterial(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Environment(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Variable(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Unit(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class SequencingTechnology(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Treatment(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Organism(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class CompoundExpression(ConfiguredBaseModel):
    
    None
    

class Measurement(CompoundExpression):
    
    value: Optional[str] = Field(None, description="""the value of the measurement""")
    unit: Optional[str] = Field(None, description="""the unit of the measurement""")
    

class CausalRelationship(CompoundExpression):
    
    cause: Optional[str] = Field(None, description="""the variable that is the cause of the effect""")
    effect: Optional[str] = Field(None, description="""the things that is affected""")
    

class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""")
    

class TextWithTriples(ConfiguredBaseModel):
    
    publication: Optional[Publication] = Field(None)
    triples: Optional[List[Triple]] = Field(default_factory=list)
    

class RelationshipType(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Publication(ConfiguredBaseModel):
    
    id: Optional[str] = Field(None, description="""The publication identifier""")
    title: Optional[str] = Field(None, description="""The title of the publication""")
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""")
    combined_text: Optional[str] = Field(None)
    full_text: Optional[str] = Field(None, description="""The full text of the publication""")
    

class AnnotatorResult(ConfiguredBaseModel):
    
    subject_text: Optional[str] = Field(None)
    object_id: Optional[str] = Field(None)
    object_text: Optional[str] = Field(None)
    


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
Study.model_rebuild()
ExtractionResult.model_rebuild()
NamedEntity.model_rebuild()
Location.model_rebuild()
EnvironmentalMaterial.model_rebuild()
Environment.model_rebuild()
Variable.model_rebuild()
Unit.model_rebuild()
SequencingTechnology.model_rebuild()
Treatment.model_rebuild()
Organism.model_rebuild()
CompoundExpression.model_rebuild()
Measurement.model_rebuild()
CausalRelationship.model_rebuild()
Triple.model_rebuild()
TextWithTriples.model_rebuild()
RelationshipType.model_rebuild()
Publication.model_rebuild()
AnnotatorResult.model_rebuild()
    



from __future__ import annotations
from datetime import datetime, date
from enum import Enum
from typing import List, Dict, Optional, Any, Union
from pydantic import BaseModel as BaseModel, Field
import sys
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


metamodel_version = "None"
version = "None"

class ConfiguredBaseModel(BaseModel,
                validate_assignment = True,
                validate_default = True,
                extra = 'forbid',
                arbitrary_types_allowed = True,
                use_enum_values = True):
    pass


class GeneLocationEnum(str, Enum):
    
    
    dummy = "dummy"
    

class GOCellComponentType(str, Enum):
    
    
    dummy = "dummy"
    

class CellType(str, Enum):
    
    
    dummy = "dummy"
    

class NullDataOptions(str, Enum):
    
    
    UNSPECIFIED_METHOD_OF_ADMINISTRATION = "UNSPECIFIED_METHOD_OF_ADMINISTRATION"
    
    NOT_APPLICABLE = "NOT_APPLICABLE"
    
    NOT_MENTIONED = "NOT_MENTIONED"
    
    

class GoCamAnnotations(ConfiguredBaseModel):
    
    genes: Optional[List[str]] = Field(default_factory=list, description="""semicolon-separated list of genes""")
    organisms: Optional[List[str]] = Field(default_factory=list, description="""semicolon-separated list of organism taxons""")
    gene_organisms: Optional[List[GeneOrganismRelationship]] = Field(default_factory=list)
    activities: Optional[List[str]] = Field(default_factory=list, description="""semicolon-separated list of molecular activities""")
    gene_functions: Optional[List[GeneMolecularActivityRelationship]] = Field(default_factory=list, description="""semicolon-separated list of gene to molecular activity relationships""")
    cellular_processes: Optional[List[str]] = Field(default_factory=list, description="""semicolon-separated list of cellular processes""")
    pathways: Optional[List[str]] = Field(default_factory=list, description="""semicolon-separated list of pathways""")
    gene_gene_interactions: Optional[List[GeneGeneInteraction]] = Field(default_factory=list, description="""semicolon-separated list of gene to gene interactions""")
    gene_localizations: Optional[List[GeneSubcellularLocalizationRelationship]] = Field(default_factory=list, description="""semicolon-separated list of genes plus their location in the cell; for example, \"gene1 / cytoplasm; gene2 / mitochondrion\"""")
    

class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    input_id: Optional[str] = Field(None)
    input_title: Optional[str] = Field(None)
    input_text: Optional[str] = Field(None)
    raw_completion_output: Optional[str] = Field(None)
    prompt: Optional[str] = Field(None)
    extracted_object: Optional[Any] = Field(None, description="""The complex objects extracted from the text""")
    named_entities: Optional[List[Any]] = Field(default_factory=list, description="""Named entities extracted from the text""")
    

class NamedEntity(ConfiguredBaseModel):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Gene(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Pathway(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class CellularProcess(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class MolecularActivity(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class GeneLocation(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Organism(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Molecule(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class CompoundExpression(ConfiguredBaseModel):
    
    None
    

class GeneOrganismRelationship(CompoundExpression):
    
    gene: Optional[str] = Field(None)
    organism: Optional[str] = Field(None)
    

class GeneMolecularActivityRelationship(CompoundExpression):
    
    gene: Optional[str] = Field(None)
    molecular_activity: Optional[str] = Field(None)
    

class GeneMolecularActivityRelationship2(CompoundExpression):
    
    gene: Optional[str] = Field(None)
    molecular_activity: Optional[str] = Field(None)
    target: Optional[str] = Field(None)
    

class GeneSubcellularLocalizationRelationship(CompoundExpression):
    
    gene: Optional[str] = Field(None)
    location: Optional[str] = Field(None)
    

class GeneGeneInteraction(CompoundExpression):
    
    gene1: Optional[str] = Field(None)
    gene2: Optional[str] = Field(None)
    

class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""")
    

class TextWithTriples(ConfiguredBaseModel):
    
    publication: Optional[Publication] = Field(None)
    triples: Optional[List[Triple]] = Field(default_factory=list)
    

class RelationshipType(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Publication(ConfiguredBaseModel):
    
    id: Optional[str] = Field(None, description="""The publication identifier""")
    title: Optional[str] = Field(None, description="""The title of the publication""")
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""")
    combined_text: Optional[str] = Field(None)
    full_text: Optional[str] = Field(None, description="""The full text of the publication""")
    

class AnnotatorResult(ConfiguredBaseModel):
    
    subject_text: Optional[str] = Field(None)
    object_id: Optional[str] = Field(None)
    object_text: Optional[str] = Field(None)
    


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
GoCamAnnotations.model_rebuild()
ExtractionResult.model_rebuild()
NamedEntity.model_rebuild()
Gene.model_rebuild()
Pathway.model_rebuild()
CellularProcess.model_rebuild()
MolecularActivity.model_rebuild()
GeneLocation.model_rebuild()
Organism.model_rebuild()
Molecule.model_rebuild()
CompoundExpression.model_rebuild()
GeneOrganismRelationship.model_rebuild()
GeneMolecularActivityRelationship.model_rebuild()
GeneMolecularActivityRelationship2.model_rebuild()
GeneSubcellularLocalizationRelationship.model_rebuild()
GeneGeneInteraction.model_rebuild()
Triple.model_rebuild()
TextWithTriples.model_rebuild()
RelationshipType.model_rebuild()
Publication.model_rebuild()
AnnotatorResult.model_rebuild()
    



from __future__ import annotations
from datetime import datetime, date
from enum import Enum
from typing import List, Dict, Optional, Any, Union
from pydantic import BaseModel as BaseModel, Field
import sys
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


metamodel_version = "None"
version = "None"

class ConfiguredBaseModel(BaseModel,
                validate_assignment = True,
                validate_default = True,
                extra = 'forbid',
                arbitrary_types_allowed = True,
                use_enum_values = True):
    pass


class NCITDrugType(str, Enum):
    
    
    dummy = "dummy"
    

class NCITTreatmentType(str, Enum):
    
    
    dummy = "dummy"
    

class NCITTActivityType(str, Enum):
    
    
    dummy = "dummy"
    

class MAXOActionType(str, Enum):
    
    
    dummy = "dummy"
    

class MESHTherapeuticType(str, Enum):
    
    
    dummy = "dummy"
    

class CHEBIDrugType(str, Enum):
    
    
    dummy = "dummy"
    

class NullDataOptions(str, Enum):
    
    
    UNSPECIFIED_METHOD_OF_ADMINISTRATION = "UNSPECIFIED_METHOD_OF_ADMINISTRATION"
    
    NOT_APPLICABLE = "NOT_APPLICABLE"
    
    NOT_MENTIONED = "NOT_MENTIONED"
    
    

class CompositeDisease(ConfiguredBaseModel):
    
    main_disease: Optional[str] = Field(None, description="""the name of the disease that is treated.""")
    drugs: Optional[List[str]] = Field(default_factory=list, description="""semicolon-separated list of named small molecule drugs""")
    treatments: Optional[List[str]] = Field(default_factory=list, description="""semicolon-separated list of therapies and treatments are indicated for treating the disease.""")
    contraindications: Optional[List[str]] = Field(default_factory=list, description="""semicolon-separated list of therapies and treatments that are contra-indicated for the disease, and should not be used, due to risk of adverse effects.""")
    treatment_mechanisms: Optional[List[TreatmentMechanism]] = Field(default_factory=list, description="""semicolon-separated list of treatment to asterisk-separated mechanism associations""")
    treatment_efficacies: Optional[List[TreatmentEfficacy]] = Field(default_factory=list, description="""semicolon-separated list of treatment to efficacy associations, e.g. Imatinib*effective""")
    treatment_adverse_effects: Optional[List[TreatmentAdverseEffect]] = Field(default_factory=list, description="""semicolon-separated list of treatment to adverse effect associations, e.g. Imatinib*nausea""")
    

class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    input_id: Optional[str] = Field(None)
    input_title: Optional[str] = Field(None)
    input_text: Optional[str] = Field(None)
    raw_completion_output: Optional[str] = Field(None)
    prompt: Optional[str] = Field(None)
    extracted_object: Optional[Any] = Field(None, description="""The complex objects extracted from the text""")
    named_entities: Optional[List[Any]] = Field(default_factory=list, description="""Named entities extracted from the text""")
    

class NamedEntity(ConfiguredBaseModel):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Gene(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Symptom(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Disease(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class AdverseEffect(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Treatment(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Mechanism(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Drug(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class CompoundExpression(ConfiguredBaseModel):
    
    None
    

class TreatmentMechanism(CompoundExpression):
    
    treatment: Optional[str] = Field(None)
    mechanism: Optional[str] = Field(None)
    

class TreatmentAdverseEffect(CompoundExpression):
    
    treatment: Optional[str] = Field(None)
    adverse_effects: Optional[List[str]] = Field(default_factory=list)
    

class TreatmentEfficacy(CompoundExpression):
    
    treatment: Optional[str] = Field(None)
    efficacy: Optional[str] = Field(None)
    

class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""")
    

class TextWithTriples(ConfiguredBaseModel):
    
    publication: Optional[Publication] = Field(None)
    triples: Optional[List[Triple]] = Field(default_factory=list)
    

class RelationshipType(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Publication(ConfiguredBaseModel):
    
    id: Optional[str] = Field(None, description="""The publication identifier""")
    title: Optional[str] = Field(None, description="""The title of the publication""")
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""")
    combined_text: Optional[str] = Field(None)
    full_text: Optional[str] = Field(None, description="""The full text of the publication""")
    

class AnnotatorResult(ConfiguredBaseModel):
    
    subject_text: Optional[str] = Field(None)
    object_id: Optional[str] = Field(None)
    object_text: Optional[str] = Field(None)
    


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
CompositeDisease.model_rebuild()
ExtractionResult.model_rebuild()
NamedEntity.model_rebuild()
Gene.model_rebuild()
Symptom.model_rebuild()
Disease.model_rebuild()
AdverseEffect.model_rebuild()
Treatment.model_rebuild()
Mechanism.model_rebuild()
Drug.model_rebuild()
CompoundExpression.model_rebuild()
TreatmentMechanism.model_rebuild()
TreatmentAdverseEffect.model_rebuild()
TreatmentEfficacy.model_rebuild()
Triple.model_rebuild()
TextWithTriples.model_rebuild()
RelationshipType.model_rebuild()
Publication.model_rebuild()
AnnotatorResult.model_rebuild()
    



"""YAML Wrapper."""
import io
import logging
from typing import Any, Optional, TextIO

import pydantic
from ruamel.yaml import YAML, RoundTripRepresenter

# import yaml
# from yaml import SafeDumper

logger = logging.getLogger(__name__)


def eliminate_empty(obj: Any, preserve=False) -> Any:
    """Eliminate empty lists and dicts from an object."""
    if isinstance(obj, list):
        return [eliminate_empty(x, preserve) for x in obj if x or preserve]
    elif isinstance(obj, dict):
        return {k: eliminate_empty(v, preserve) for k, v in obj.items() if v or preserve}
    elif isinstance(obj, pydantic.BaseModel):
        return eliminate_empty(obj.model_dump(), preserve)
    elif isinstance(obj, tuple):
        return [eliminate_empty(x, preserve) for x in obj]
    elif isinstance(obj, str):
        return str(obj)
    else:
        return obj


# SafeDumper.add_representer(str,
#     lambda dumper, value: dumper.represent_scalar('tag:yaml.org,2002:str', value, style='|')
# )


def repr_str(dumper: RoundTripRepresenter, data: str):
    if "\n" in data:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


def dump_minimal_yaml(obj: Any, minimize=True, file: Optional[TextIO] = None) -> str:
    """Dump a YAML string, but eliminating Nones and empty lists and dicts."""
    yaml = YAML()
    yaml.representer.add_representer(str, repr_str)
    yaml.default_flow_style = False
    yaml.indent(sequence=4, offset=2)
    if not file:
        file = io.StringIO()
        yaml.dump(eliminate_empty(obj, not minimize), file)
        return file.getvalue()
    else:
        yaml.dump(eliminate_empty(obj, not minimize), file)
        return ""



from __future__ import annotations
from datetime import datetime, date
from enum import Enum
from typing import List, Dict, Optional, Any, Union
from pydantic import BaseModel as BaseModel, Field
import sys
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


metamodel_version = "None"
version = "None"

class ConfiguredBaseModel(BaseModel,
                validate_assignment = True,
                validate_default = True,
                extra = 'forbid',
                arbitrary_types_allowed = True,
                use_enum_values = True):
    pass


class NullDataOptions(str, Enum):
    
    
    UNSPECIFIED_METHOD_OF_ADMINISTRATION = "UNSPECIFIED_METHOD_OF_ADMINISTRATION"
    
    NOT_APPLICABLE = "NOT_APPLICABLE"
    
    NOT_MENTIONED = "NOT_MENTIONED"
    
    

class IBDAnnotations(ConfiguredBaseModel):
    
    genes: Optional[List[str]] = Field(default_factory=list, description="""semicolon-separated list of genes""")
    exposures: Optional[List[str]] = Field(default_factory=list, description="""semicolon-separated list of exposures""")
    gene_exposures_relationships: Optional[List[GeneExposureRelationship]] = Field(default_factory=list, description="""semicolon-separated list of gene to molecular activity relationships""")
    diseases: Optional[List[str]] = Field(default_factory=list, description="""semicolon-separated list of diseases""")
    cellular_process: Optional[List[str]] = Field(default_factory=list, description="""semicolon-separated list of cellular processes""")
    disease_cellular_process_relationships: Optional[List[DiseaseCellularProcessRelationship]] = Field(default_factory=list, description="""semicolon-separated list of disease to cellular process relationships""")
    

class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    input_id: Optional[str] = Field(None)
    input_title: Optional[str] = Field(None)
    input_text: Optional[str] = Field(None)
    raw_completion_output: Optional[str] = Field(None)
    prompt: Optional[str] = Field(None)
    extracted_object: Optional[Any] = Field(None, description="""The complex objects extracted from the text""")
    named_entities: Optional[List[Any]] = Field(default_factory=list, description="""Named entities extracted from the text""")
    

class NamedEntity(ConfiguredBaseModel):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Gene(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class ChemicalExposure(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class ChemicalExposureToGenePredicate(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Disease(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class CellularProcess(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class DiseaseToCellularProcessPredicate(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class CompoundExpression(ConfiguredBaseModel):
    
    None
    

class GeneExposureRelationship(CompoundExpression):
    
    subject: Optional[str] = Field(None, description="""The name of the exposure, such as a exposure to a chemical toxin.""")
    predicate: Optional[str] = Field(None, description="""The name of the type of relationship between a chemical exposure and a gene.""")
    object: Optional[str] = Field(None, description="""The name of the gene in the pair. This comes second in the pair.""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the chemical exposure.""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the gene.""")
    

class DiseaseCellularProcessRelationship(CompoundExpression):
    
    subject: Optional[str] = Field(None, description="""The name of the disease.""")
    predicate: Optional[str] = Field(None, description="""The name of the type of relationship between a disease and a cellular process.""")
    object: Optional[str] = Field(None, description="""The name of the cellular process.""")
    subject_qualifier: Optional[str] = Field(None)
    object_qualifier: Optional[str] = Field(None)
    

class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""")
    

class TextWithTriples(ConfiguredBaseModel):
    
    publication: Optional[Publication] = Field(None)
    triples: Optional[List[Triple]] = Field(default_factory=list)
    

class RelationshipType(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Publication(ConfiguredBaseModel):
    
    id: Optional[str] = Field(None, description="""The publication identifier""")
    title: Optional[str] = Field(None, description="""The title of the publication""")
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""")
    combined_text: Optional[str] = Field(None)
    full_text: Optional[str] = Field(None, description="""The full text of the publication""")
    

class AnnotatorResult(ConfiguredBaseModel):
    
    subject_text: Optional[str] = Field(None)
    object_id: Optional[str] = Field(None)
    object_text: Optional[str] = Field(None)
    


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
IBDAnnotations.model_rebuild()
ExtractionResult.model_rebuild()
NamedEntity.model_rebuild()
Gene.model_rebuild()
ChemicalExposure.model_rebuild()
ChemicalExposureToGenePredicate.model_rebuild()
Disease.model_rebuild()
CellularProcess.model_rebuild()
DiseaseToCellularProcessPredicate.model_rebuild()
CompoundExpression.model_rebuild()
GeneExposureRelationship.model_rebuild()
DiseaseCellularProcessRelationship.model_rebuild()
Triple.model_rebuild()
TextWithTriples.model_rebuild()
RelationshipType.model_rebuild()
Publication.model_rebuild()
AnnotatorResult.model_rebuild()
    



"""Utilities for parsing text."""
from typing import List


def split_on_one_of(text: str, separators: List[str]) -> List[str]:
    """Split text on the first separator found."""
    for sep in separators:
        if sep in text:
            return text.split(sep)
    return [text]



"""Tools to extract sub-ontologies and reasoner tasks."""
import base64
import logging
import random
import re
import sys
import uuid
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    ClassVar,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    TextIO,
    Tuple,
    Type,
    Union,
)

import inflection
import yaml
from oaklib.datamodels.taxon_constraints import SubjectTerm, Taxon, TaxonConstraint
from oaklib.datamodels.vocabulary import (
    DISJOINT_WITH,
    IN_TAXON,
    INVERSE_OF,
    IS_A,
    NEVER_IN_TAXON,
    ONLY_IN_TAXON,
    OWL_CLASS,
    OWL_NAMED_INDIVIDUAL,
    OWL_SYMMETRIC_PROPERTY,
    OWL_TRANSITIVE_PROPERTY,
    PART_OF,
    SUBPROPERTY_OF,
)
from oaklib.implementations import SqlImplementation
from oaklib.interfaces import OboGraphInterface
from oaklib.interfaces.basic_ontology_interface import RELATIONSHIP
from oaklib.interfaces.obograph_interface import GraphTraversalMethod
from oaklib.interfaces.semsim_interface import SemanticSimilarityInterface
from oaklib.interfaces.taxon_constraint_interface import TAXON_PREDICATES, TaxonConstraintInterface
from oaklib.types import CURIE, PRED_CURIE
from oaklib.utilities.obograph_utils import shortest_paths
from pydantic import BaseModel, Field
from semsql.sqla.semsql import RdfListMemberStatement, RdfTypeStatement, Statements

logger = logging.getLogger(__name__)


OBFUSCATED_ID = "str"


class Axiom(BaseModel):
    """Represents an individual logical axiom."""

    text: str = ""
    """Textual representation of the axiom, e.g A SubClassOf B"""

    obfuscated_text: Optional[str] = None
    """Textual representation of the axiom, with obfuscated terms"""

    abox: bool = False
    """Whether the axiom is an ABox axiom."""

    edge: Optional[RELATIONSHIP] = None
    """The relationship between the two terms in the axiom."""


class Ontology(BaseModel):
    """An ontology is a collection of axioms."""

    name: Optional[str] = None
    axioms: List[Axiom]
    """All axioms in the ontology"""

    terms: List[CURIE] = [""]
    predicates: List[PRED_CURIE] = [""]

    comments: Optional[List[str]] = None


class Query(BaseModel):
    """Query."""

    text: Optional[str] = None
    """Text representation of the query."""

    parameters: Optional[List[str]] = None
    """Parameters to the query.
    E.g for a query of the form 'what is the superclass of X',
    there is one parameter X, for the subclass."""


class Explanation(BaseModel):
    """A set of axioms that entail some explained axiom."""

    axioms: List[Axiom] = []
    text: Optional[str] = None
    comments: Optional[List[str]] = None


class Answer(BaseModel):
    """Individual answer to a query."""

    _value_domain: ClassVar[str]
    """Class variable indicating answer types."""

    text: str
    """Textual representation of the answer."""

    explanations: Optional[List[Explanation]] = None
    """All explanations for the answer."""

    def shortest_explanation(self) -> Explanation:
        """Return the shortest explanation for the answer."""
        if not self.explanations:
            return Explanation(axioms=[Axiom(text="No explanation found")])
        shortest = min(self.explanations, key=lambda x: len(x.axioms))
        return shortest


class ObjectAnswer(Answer):
    """Answer that is an object, e.g class."""

    _value_domain = "The name of the object."


class ClassAnswer(Answer):
    """Answer that is an OWL class."""

    _value_domain = "The name of the class."


class InstanceAnswer(Answer):
    """Answer that is an OWL individual."""

    _value_domain = "The name of the individual."


class BooleanAnswer(Answer):
    """Answer that is a boolean, e.g. true or false."""

    _value_domain = "Either TRUE or FALSE."


class AxiomAnswer(Answer):
    """Answer that is a axiom."""

    _value_domain = "An axiom written in Manchester syntax."


class ExampleQueryAnswers(BaseModel):
    """An example query, plus all expected answers."""

    query: Query
    answers: Optional[List[Answer]] = None


class Example(BaseModel):
    """An example of a query plus answers, in the context of an ontology."""

    ontology: Ontology
    query_answers: Optional[List[ExampleQueryAnswers]] = None


class GPTReasonMethodType(str, Enum):
    BASIC = "basic"
    EXPLANATION = "explanation"
    CHAIN_OF_THOUGHT = "chain_of_thought"


class Task(BaseModel):
    """
    A task is a query on an ontology that has a set of defined answers.

    For example, a task group might be to determine if an ontology is consistent.
    """

    _query_format: ClassVar[str]

    type: Literal["Task"] = Field("Task")
    _code: ClassVar[str]

    has_multiple_answers: ClassVar[bool] = True

    ontology: Ontology
    name: Optional[str] = None
    description: Optional[str] = None
    query: Optional[Query] = None
    answers: Optional[List[Answer]] = None
    examples: Optional[List[Example]] = None
    obfuscated: Optional[bool] = False

    method: Optional[GPTReasonMethodType] = None

    include_explanations: Optional[bool] = False
    """If true then completing the task must involve providing explanations for each answer."""

    chain_of_thought: Optional[bool] = False
    """If true explanations come first, providing chain of thought reasoning."""

    abductive: Optional[bool] = False
    """If true then the task is to find explanations for answers that are given."""

    shortest_explanation: Optional[Explanation] = None

    len_shortest_explanation: Optional[int] = None

    class Config:
        """Pydantic configuration."""

        use_enum_values = True

    def populate(self) -> None:
        qf = self._query_format
        for example in self.examples:
            for query_answer in example.query_answers:
                if not query_answer.query.text:
                    query_answer.query.text = qf.format(params=query_answer.query.parameters)
        if self.query is not None:
            if not self.query.text:
                self.query.text = qf.format(params=self.query.parameters)
        if self.answers is not None:
            if len(self.answers) == 0:
                self.shortest_explanation = None
                self.len_shortest_explanation = 0
            else:
                most_complex_answer = max(
                    self.answers, key=lambda x: len(x.shortest_explanation().axioms)
                )
                if self.shortest_explanation is not None:
                    self.shortest_explanation = most_complex_answer.shortest_explanation()
                    self.len_shortest_explanation = len(self.shortest_explanation.axioms)
        if not self.name:
            self.name = f"{self.type}-{uuid.uuid4()}"
        self.init_method()

    def init_method(self):
        if self.method:
            logger.info(f"Initializing method for {self.name}")
            if not isinstance(self.method, GPTReasonMethodType):
                self.method = GPTReasonMethodType(self.method)
            if self.method == GPTReasonMethodType.EXPLANATION:
                self.include_explanations = True
            elif self.method == GPTReasonMethodType.CHAIN_OF_THOUGHT:
                self.chain_of_thought = True
        else:
            if self.include_explanations:
                self.method = GPTReasonMethodType.EXPLANATION
            elif self.chain_of_thought:
                self.method = GPTReasonMethodType.CHAIN_OF_THOUGHT
            else:
                self.method = GPTReasonMethodType.BASIC


class OntologyCoherencyTask(Task):
    """
    A task to determine if an ontology is coherent.

    There should be a single answer, which is a boolean.
    """

    _query_format = """
    What are the unsatisfiable classes in the ontology?
    A class X is unsatisfiable if X is a subclass of DC1 and also a subclass of DC2,
    where there is an axiom DC1 DisjointWith DC2.
    Here, SubClassOf can mean either direct or indirect subclass, entailed
    through the transitivity of SubClassOf.
    List all unsatisfiable classes that can be found with this rule.
    If there are no unsatisfiable classes, just write NONE."""

    type: Literal["OntologyCoherencyTask"] = Field("OntologyCoherencyTask")  # type: ignore
    _code: str = "sat"  # type: ignore

    has_multiple_answers = False
    answers: Optional[List[ClassAnswer]] = None

    examples: Optional[List[Example]] = [
        Example(
            ontology=Ontology(
                axioms=[
                    Axiom(text="B SubClassOf A"),
                    Axiom(text="C SubClassOf A"),
                    Axiom(text="D SubClassOf B"),
                    Axiom(text="D SubClassOf C"),
                    Axiom(text="B DisjointWith C"),
                ]
            ),
            query_answers=[
                ExampleQueryAnswers(
                    query=Query(),
                    answers=[
                        ClassAnswer(
                            text="D",
                            explanations=[
                                Explanation(
                                    text="""D is unsatisfiable because it is a subclass
                                            of two disjoint classes (B and C).""",
                                    axioms=[
                                        Axiom(text="D SubClassOf B"),
                                        Axiom(text="D SubClassOf C"),
                                        Axiom(text="B DisjointWith C"),
                                    ],
                                )
                            ],
                        )
                    ],
                ),
            ],
        ),
        Example(
            ontology=Ontology(
                axioms=[
                    Axiom(text="B SubClassOf A"),
                    Axiom(text="C SubClassOf A"),
                    Axiom(text="D SubClassOf B"),
                    Axiom(text="B DisjointWith A"),
                ]
            ),
            query_answers=[
                ExampleQueryAnswers(
                    query=Query(),
                    answers=[
                        ClassAnswer(
                            text="NONE",
                            explanations=[
                                Explanation(
                                    text="""There is only one Disjointness Axiom (B DisjointWith A),
                                    and B and A do not share either direct
                                    or indirect subclasses.""",
                                    axioms=[],
                                )
                            ],
                        )
                    ],
                ),
            ],
        ),
    ]


class EntailedIndirectSuperClassTask(Task):
    """A task to determine the indirect superclasses of a class."""

    _query_format = """
    What are the indirect entailed superclasses of {params[0]}?
    Include answers entailed by the transitivity of SubClassOf.
    Do not include direct (one-hop) superclasses.
    """

    type: Literal["EntailedIndirectSuperClassTask"] = Field(
        "EntailedIndirectSuperClassTask"
    )  # type: ignore
    _code: str = "indirect"  # type: ignore

    answers: Optional[List[ClassAnswer]] = None

    examples: Optional[List[Example]] = [
        Example(
            ontology=Ontology(
                axioms=[
                    Axiom(text="E2 SubClassOf E"),
                    Axiom(text="E SubClassOf B"),
                    Axiom(text="B SubClassOf A"),
                    Axiom(text="C SubClassOf A"),
                    Axiom(text="D SubClassOf B"),
                ]
            ),
            query_answers=[
                ExampleQueryAnswers(
                    query=Query(parameters=["E"]),
                    answers=[
                        ClassAnswer(
                            text="A",
                            explanations=[
                                Explanation(
                                    text="""A is an indirect entailed superclass of E because
                                    E SubClassOf B, and B SubClassOf A, and SubClassOf is
                                    transitive.""",
                                    axioms=[
                                        Axiom(text="E SubClassOf B"),
                                        Axiom(text="B SubClassOf A"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
                ExampleQueryAnswers(
                    query=Query(parameters=["E2"]),
                    answers=[
                        ClassAnswer(
                            text="A",
                            explanations=[
                                Explanation(
                                    text="""A is an indirect entailed superclass of E2 because
                                    E2 SubClassOf E, and E SubClassOf B, and B SubClassOf A,
                                    and because SubClassOf is transitive.""",
                                    axioms=[
                                        Axiom(text="E2 SubClassOf E"),
                                        Axiom(text="E SubClassOf B"),
                                        Axiom(text="B SubClassOf A"),
                                    ],
                                )
                            ],
                        ),
                        ClassAnswer(
                            text="B",
                            explanations=[
                                Explanation(
                                    text="""B is an indirect entailed superclass of E2 because
                                    E2 SubClassOf E, and E SubClassOf B, and because SubClassOf
                                    is transitive.""",
                                    axioms=[
                                        Axiom(text="E SubClassOf B"),
                                        Axiom(text="E2 SubClassOf E"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
            ],
        )
    ]


class EntailedTransitiveSuperClassTask(Task):
    """A task to determine the all transitive superclasses of a class."""

    _query_format = """
    What are the transitive superclasses of {params[0]}?
    Include answers entailed by the transitivity of SubClassOf.
    Also direct (one-hop) superclasses.
    """

    type: Literal["EntailedTransitiveSuperClassTask"] = Field(
        "EntailedTransitiveSuperClassTask"
    )  # type: ignore
    _code: str = "superc"  # type: ignore

    answers: Optional[List[ClassAnswer]] = None

    examples: Optional[List[Example]] = [
        Example(
            ontology=Ontology(
                axioms=[
                    Axiom(text="E2 SubClassOf E"),
                    Axiom(text="E SubClassOf B"),
                    Axiom(text="B SubClassOf A"),
                    Axiom(text="C SubClassOf A"),
                    Axiom(text="D SubClassOf B"),
                ]
            ),
            query_answers=[
                ExampleQueryAnswers(
                    query=Query(parameters=["E"]),
                    answers=[
                        ClassAnswer(
                            text="A",
                            explanations=[
                                Explanation(
                                    text="""A is an entailed superclass of E because
                                    E SubClassOf B, and B SubClassOf A, and SubClassOf is
                                    transitive.""",
                                    axioms=[
                                        Axiom(text="E SubClassOf B"),
                                        Axiom(text="B SubClassOf A"),
                                    ],
                                )
                            ],
                        ),
                        ClassAnswer(
                            text="B",
                            explanations=[
                                Explanation(
                                    text="""B is an indirect entailed superclass of E because
                                        it is already asserted.""",
                                    axioms=[
                                        Axiom(text="B SubClassOf A"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
                ExampleQueryAnswers(
                    query=Query(parameters=["E2"]),
                    answers=[
                        ClassAnswer(
                            text="A",
                            explanations=[
                                Explanation(
                                    text="""A is an indirect entailed superclass of E2 because
                                    E2 SubClassOf E, and E SubClassOf B, and B SubClassOf A,
                                    and because SubClassOf is transitive.""",
                                    axioms=[
                                        Axiom(text="E2 SubClassOf E"),
                                        Axiom(text="E SubClassOf B"),
                                        Axiom(text="B SubClassOf A"),
                                    ],
                                )
                            ],
                        ),
                        ClassAnswer(
                            text="B",
                            explanations=[
                                Explanation(
                                    text="""B is an entailed superclass of E2 because
                                    E2 SubClassOf E, and E SubClassOf B, and because SubClassOf
                                    is transitive.""",
                                    axioms=[
                                        Axiom(text="E SubClassOf B"),
                                        Axiom(text="E2 SubClassOf E"),
                                    ],
                                )
                            ],
                        ),
                        ClassAnswer(
                            text="E",
                            explanations=[
                                Explanation(
                                    text="""E is an entailed superclass of E2 because
                                    it is directly asserted.""",
                                    axioms=[
                                        Axiom(text="E2 SubClassOf E"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
            ],
        )
    ]


class EntailedSubClassOfExpressionTask(Task):
    """A task to determine the subclasses of a class expression."""

    _query_format = """
    What are the entailed subclasses of the expression {params[0]} Some {params[1]}?
    Include indirect (transitive) descendants.
    """

    type: Literal["EntailedSubClassOfExpressionTask"] = Field(
        "EntailedSubClassOfExpressionTask"
    )  # type: ignore
    _code: str = "expr"  # type: ignore

    answers: Optional[List[ClassAnswer]] = None

    examples: Optional[List[Example]] = [
        Example(
            ontology=Ontology(
                axioms=[
                    Axiom(text="P type TransitiveProperty"),
                    Axiom(text="E2 SubClassOf P some E"),
                    Axiom(text="E SubClassOf B"),
                    Axiom(text="B SubClassOf P some A"),
                    Axiom(text="C SubClassOf A"),
                    Axiom(text="D SubClassOf Q some B"),
                ]
            ),
            query_answers=[
                ExampleQueryAnswers(
                    query=Query(parameters=["P", "A"]),
                    answers=[
                        ClassAnswer(
                            text="B",
                            explanations=[
                                Explanation(
                                    text="""B is an entailed subclass of P some A because
                                    it is directly asserted in the ontology.""",
                                    axioms=[
                                        Axiom(text="B SubClassOf P some A"),
                                    ],
                                )
                            ],
                        ),
                        ClassAnswer(
                            text="E",
                            explanations=[
                                Explanation(
                                    text="""E is an entailed subclass of P some A because
                                    E SubClassOf B and B SubClassOf P some A
                                    (i.e. every B stands in relation P to some A)""",
                                    axioms=[
                                        Axiom(text="E SubClassOf B"),
                                        Axiom(text="B SubClassOf P some A"),
                                    ],
                                )
                            ],
                        ),
                        ClassAnswer(
                            text="E2",
                            explanations=[
                                Explanation(
                                    text="""E2 is an entailed subclass of P some A because
                                    E2 SubClassOf E and E SubClassOf B and B SubClassOf P some A
                                    (i.e. every B stands in relation P to some A)""",
                                    axioms=[
                                        Axiom(text="B SubClassOf P some A"),
                                        Axiom(text="E SubClassOf B"),
                                        Axiom(text="E2 SubClassOf P some E"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
                ExampleQueryAnswers(
                    query=Query(parameters=["Q", "B"]),
                    answers=[
                        ClassAnswer(
                            text="D",
                            explanations=[
                                Explanation(
                                    text="""D is an entailed subclass of Q some B because
                                    it is directly asserted in the ontology. This is the only
                                    answer because Q is not declared transitive.""",
                                    axioms=[
                                        Axiom(text="D SubClassOf Q some B"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
            ],
        )
    ]


class EntailedDirectSuperClassTask(Task):
    """
    A task to determine the direct superclasses of a class.

    Includes those entailed by other axioms, e.g. equivalence axioms.

    Context: a standard pattern in bio-ontologies is to infer the structure
    of one ontology from another - e.g. the metabolic process branch in GO
    may be entailed by GO equivalence axioms plus the IS_A links in CHEBI.
    """

    _query_format = """
    What are the direct entailed superclasses of {params[0]}?.
    Make use of all axioms in the provided ontology.
    """

    type: Literal["EntailedDirectSuperClassTask"] = Field(
        "EntailedDirectSuperClassTask"
    )  # type: ignore
    _code: str = "dir-sup"  # type: ignore

    answers: Optional[List[ClassAnswer]] = None

    # TODO: examples


class MostRecentCommonSubsumerTask(Task):
    """A task to determine the most specific common ancestors."""

    _query_format = """
    What are the most specific common entailed superclasses of {params[0]} and {params[1]}?.
    """

    type: Literal["MostRecentCommonSubsumerTask"] = Field(
        "MostRecentCommonSubsumerTask"
    )  # type: ignore
    _code: str = "mrca"  # type: ignore

    answers: Optional[List[ClassAnswer]] = None

    examples: Optional[List[Example]] = [
        Example(
            ontology=Ontology(
                axioms=[
                    Axiom(text="E2 SubClassOf E"),
                    Axiom(text="E SubClassOf B"),
                    Axiom(text="B SubClassOf A"),
                    Axiom(text="C SubClassOf A"),
                    Axiom(text="D SubClassOf B"),
                ]
            ),
            query_answers=[
                ExampleQueryAnswers(
                    query=Query(parameters=["E2", "D"]),
                    answers=[
                        ClassAnswer(
                            text="B",
                            explanations=[
                                Explanation(
                                    text="""B is the most specific common entailed superclass
                                    of E2 and D because E2 is a SubClassOf B via E, and D is
                                    a direct SubClassOf B, and there are no more specific common
                                    ancestors.""",
                                    axioms=[
                                        Axiom(text="E2 SubClassOf E"),
                                        Axiom(text="E SubClassOf B"),
                                        Axiom(text="D SubClassOf B"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
                ExampleQueryAnswers(
                    query=Query(parameters=["E2", "C"]),
                    answers=[
                        ClassAnswer(
                            text="A",
                            explanations=[
                                Explanation(
                                    text="""A is the most specific common entailed superclass
                                    of E2 and C because E2 is a SubClassOf A via E then B, and C is
                                    a SubClassOf A via B, and there are no more specific common
                                    ancestors.""",
                                    axioms=[
                                        Axiom(text="E2 SubClassOf E"),
                                        Axiom(text="E SubClassOf B"),
                                        Axiom(text="B SubClassOf A"),
                                        Axiom(text="C SubClassOf B"),
                                        Axiom(text="B SubClassOf A"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
                ExampleQueryAnswers(
                    query=Query(parameters=["E2", "E"]),
                    answers=[
                        ClassAnswer(
                            text="E",
                            explanations=[
                                Explanation(
                                    text="""E is the most specific common entailed superclass of E2 and E because
                                    trivially E2 SubClassOf E""",
                                    axioms=[
                                        Axiom(text="E2 SubClassOf E"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
            ],
        )
    ]


class TaxonConstraintTask(Task):
    """A task to determine inapplicable classes by taxon."""

    _query_format = """
    Is the term {params[0]} applicable for the taxon {params[1]}?.
    Make use of subclass and part of axioms.
    subclass is transitive: if A subClassOf B and B subClassOf C, then A subClassOf C.
    part of is transitive: if A subClassOf part_of some B, and B subClassOf part_of some C,
    then A subClassOf part_of some C.
    subclass and part of chain: if A subClassOf B and B subClassOf part_of some C,
    then A subClassOf part_of some C.
    part of and subclass chain: if A subClassOf part_of some B and B subClassOf C,
    then A subClassOf part_of some C.
    only_in_taxon restricts a term to a taxon: if A only_in_taxon T,
    and T DisjointWith T2, then A is INVALID for T2.
    never_in_taxon restricts a term to a taxon: if A never_in_taxon T,
    then A is INVALID for T.
    only_in_taxon propagates up subClassOf: if A only_in_taxon T, and T subClassOf T2,
    then A only_in_taxon T2.
    never_in_taxon propagates down subClassOf: if A only_in_taxon T, and T2 subClassOf T,
    then A never_in_taxon T2.
    """

    type: Literal["TaxonConstraintTask"] = Field("TaxonConstraintTask")  # type: ignore
    _code: str = "tc"  # type: ignore

    answers: Optional[List[ClassAnswer]] = None

    examples: Optional[List[Example]] = [
        Example(
            ontology=Ontology(
                axioms=[
                    Axiom(text="E2 SubClassOf part_of some E"),
                    Axiom(text="E SubClassOf B"),
                    Axiom(text="B SubClassOf A"),
                    Axiom(text="C SubClassOf A"),
                    Axiom(text="D SubClassOf B"),
                    Axiom(text="D1 SubClassOf D"),
                    Axiom(text="D2 SubClassOf D"),
                    Axiom(text="Tax1 SubClassOf Tax1Root"),
                    Axiom(text="Tax2 SubClassOf Tax1Root"),
                    Axiom(text="Tax1 DisjointWith Tax2"),
                    Axiom(text="Tax1x SubClassOf Tax1"),
                    Axiom(text="Tax1y SubClassOf Tax1"),
                    Axiom(text="Tax1x DisjointWith Tax1y"),
                    Axiom(text="Tax1xF SubClassOf Tax1x"),
                    Axiom(text="Tax1xG SubClassOf Tax1x"),
                    Axiom(text="Tax1xF DisjointWith Tax1xG"),
                    Axiom(text="Tax1yJ SubClassOf Tax1y"),
                    Axiom(text="Tax1yK SubClassOf Tax1y"),
                    Axiom(text="Tax1yJ DisjointWith Tax1yK"),
                    Axiom(text="E only_in_taxon Tax1"),
                    Axiom(text="E never_in_taxon Tax1x"),
                    Axiom(text="D only_in_taxon Tax1y"),
                ]
            ),
            query_answers=[
                ExampleQueryAnswers(
                    query=Query(parameters=["E2", "Tax2"]),
                    answers=[
                        ClassAnswer(
                            text="INVALID",
                            explanations=[
                                Explanation(
                                    text="""Every E2 is part of some E,
                                            and E is only found in Tax1,
                                            and Tax1 is disjoint with Tax2,
                                            therefore E2 is invalid for Tax2""",
                                    axioms=[
                                        Axiom(text="E2 SubClassOf part_of some E"),
                                        Axiom(text="E only_in_taxon Tax1"),
                                        Axiom(text="Tax1 DisjointWith Tax2"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
                ExampleQueryAnswers(
                    query=Query(parameters=["E2", "Tax1y"]),
                    answers=[
                        ClassAnswer(
                            text="VALID",
                            explanations=[
                                Explanation(
                                    text="""Every E2 is part of some E,
                                            E SubClassOf B, B SubClassOf A,
                                            and there are no conflicting taxon
                                            constraints for E2, E, B, or A,
                                            therefore E2 is valid for Tax1y""",
                                    axioms=[
                                        Axiom(text="E2 SubClassOf part_of some E"),
                                        Axiom(text="E SubClassOf B"),
                                        Axiom(text="B SubClassOf A"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
                ExampleQueryAnswers(
                    query=Query(parameters=["D", "Tax1y"]),
                    answers=[
                        ClassAnswer(
                            text="VALID",
                            explanations=[
                                Explanation(
                                    text="""Every D is part of some B, B SubClassOf A,
                                            and there are no conflicting taxon
                                            constraints for D, B, or A,
                                            therefore D is valid for Tax1y""",
                                    axioms=[
                                        Axiom(text="D SubClassOf part_of some B"),
                                        Axiom(text="B SubClassOf A"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
                ExampleQueryAnswers(
                    query=Query(parameters=["E2", "Tax1xF"]),
                    answers=[
                        ClassAnswer(
                            text="INVALID",
                            explanations=[
                                Explanation(
                                    text="""Every E2 is part of some E, and E is never
                                            found in Tax1x,
                                            Tax1xF is a subclass of Tax1x,
                                            therefore E2 is invalid for Tax1xF""",
                                    axioms=[
                                        Axiom(text="E2 SubClassOf part_of some E"),
                                        Axiom(text="E never_in_taxon Tax1x"),
                                        Axiom(text="Tax1xF SubClassOf Tax1x"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
            ],
        )
    ]


class ABoxTask(Task):
    """A task to infer assertions over property chains and transitvity in aboxes."""

    _query_format = """
    What instances <I> satisfy {params[0]} {params[1]} <I> ?.
    Make use of property chain axioms of the form
    PROPERTY1 o PROPERTY2 SubPropertyOf PROPERTY3.
    This means that if x PROPERTY1 y and y PROPERTY2 z then x PROPERTY3 z.
    Also make use of transitivity axioms of the form
    PROPERTY type TransitiveProperty.
    This means that if x PROPERTY y and y PROPERTY z then x PROPERTY z.
    """

    type: Literal["ABoxTask"] = Field("ABoxTask")  # type: ignore
    _code: str = "abox"  # type: ignore

    answers: Optional[List[InstanceAnswer]] = None

    # TODO: examples

    examples: Optional[List[Example]] = [
        Example(
            ontology=Ontology(
                axioms=[
                    Axiom(text="p1 o p2 SubPropertyOf p3"),
                    Axiom(text="p1 type TransitiveProperty"),
                    Axiom(text="i0 p1 i1"),
                    Axiom(text="i1 p1 i2"),
                    Axiom(text="i2 p2 i3"),
                    Axiom(text="i3 p1 i4"),
                ],
                comments=["""a chain of two transitive properties followed by a property chain."""],
            ),
            query_answers=[
                ExampleQueryAnswers(
                    query=Query(parameters=["i0", "p3"]),
                    answers=[
                        InstanceAnswer(
                            text="i3",
                            explanations=[
                                Explanation(
                                    text="""i0 p3 i3 because
                                    i0 p1 i1 and i1 p1 i2 and p1 is transitive, so i0 p1 i2.
                                    i2 p2 i3 and p1 o p2 SubPropertyOf p3, so i0 p3 i3""",
                                    axioms=[
                                        Axiom(text="i0 p1 i1"),
                                        Axiom(text="i1 p1 i2"),
                                        Axiom(text="p1 type TransitiveProperty"),
                                        Axiom(text="i2 p2 i3"),
                                        Axiom(text="p1 o p2 SubPropertyOf p3"),
                                    ],
                                )
                            ],
                        ),
                        InstanceAnswer(
                            text="i2",
                            explanations=[
                                Explanation(
                                    text="""i0 p3 i2 because
                                        i0 p1 i1 and i1 p1 i2 and p1 is transitive, so i0 p1 i2.""",
                                    axioms=[
                                        Axiom(text="i0 p1 i1"),
                                        Axiom(text="i1 p1 i2"),
                                        Axiom(text="p1 type TransitiveProperty"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
                ExampleQueryAnswers(
                    query=Query(parameters=["i1", "p3"]),
                    answers=[
                        InstanceAnswer(
                            text="i3",
                            explanations=[
                                Explanation(
                                    text="""i1 p3 i3 because
                                            i1 p1 i2 and
                                            i2 p2 i3 and p1 o p2 SubPropertyOf p3, so i1 p3 i3""",
                                    axioms=[
                                        Axiom(text="i1 p1 i2"),
                                        Axiom(text="i2 p2 i3"),
                                        Axiom(text="p1 o p2 SubPropertyOf p3"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
                ExampleQueryAnswers(
                    query=Query(parameters=["i0", "p1"]),
                    answers=[
                        InstanceAnswer(
                            text="i1",
                            explanations=[
                                Explanation(
                                    text="""i0 p1 i1 is directly asserted""",
                                    axioms=[
                                        Axiom(text="i0 p1 i1"),
                                    ],
                                )
                            ],
                        ),
                        InstanceAnswer(
                            text="i2",
                            explanations=[
                                Explanation(
                                    text="""i0 p1 i2 because
                                            i0 p1 i1 and i1 p1 i2 and p1 is transitive,
                                            so i0 p1 i2.""",
                                    axioms=[
                                        Axiom(text="i0 p1 i1"),
                                        Axiom(text="i1 p1 i2"),
                                        Axiom(text="p1 type TransitiveProperty"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
            ],
        )
    ]


class TaskCollection(BaseModel):
    name: Optional[str] = None
    tasks: List[Task] = None
    obfuscated_curie_map: Optional[Mapping[str, str]] = None
    name_map: Optional[Mapping[str, str]] = None

    @staticmethod
    def load(file_or_object: Union[dict, str, Path, TextIO]):
        if isinstance(file_or_object, Path):
            file_or_object = str(file_or_object)
        if isinstance(file_or_object, str):
            with open(file_or_object) as f:
                tc_dict = yaml.safe_load(f)
        else:
            tc_dict = yaml.safe_load(str(file_or_object))
        current_module = sys.modules[__name__]
        tasks = []
        for task_dict in tc_dict["tasks"]:
            typ = task_dict["type"]
            cls = current_module.__dict__[typ]
            task = cls(**task_dict)
            if not isinstance(task.method, GPTReasonMethodType):
                # TODO: figure how to get pydantic to do this
                task.method = GPTReasonMethodType(task.method)
            tasks.append(task)
        tc_dict["tasks"] = tasks
        return TaskCollection(**tc_dict)


@dataclass
class OntologyExtractor:
    """
    Ontology extractor.

    This will extract Task objects from an ontology. These Task objects
    can be used by a ReasonerEngine object as queries or evaluation tests.

    Examples of types of Task objects include:

    - OntologyCoherencyTask: determine if an ontology is coherent
    - EntailedIndirectSuperClassTask: determine the indirect superclasses of a class

    The extraction process is guided by seed classes or objects, and the resulting Task
    object contains a minimal ontology (similar to a SLME module) that should be sufficient
    to answer the query, as well as a specification of the task.
    """

    adapter: OboGraphInterface
    additional_axioms: Optional[List[Axiom]] = None
    query_predicates: Optional[List[PRED_CURIE]] = None

    use_identifiers: bool = False
    name_map: Optional[Mapping[str, str]] = field(default_factory=lambda: {})  # type: ignore
    obfuscate: bool = False
    obfuscated_curie_map: Optional[Mapping[str, str]] = field(
        default_factory=lambda: {}  # type: ignore
    )

    def create_task(
        self, task_type: Union[str, Type[Task]], parameters: Optional[List[Any]] = None, **kwargs
    ) -> Task:
        terms = list(self.adapter.entities(filter_obsoletes=True, owl_type=OWL_CLASS))
        ontology = self.extract_ontology(terms)
        if isinstance(task_type, str):
            task_type = globals()[task_type]
        if not isinstance(task_type, str):  # Just to be sure!
            task = task_type(ontology=ontology, query=Query(parameters=parameters))
        task.populate()
        return task

    def create_random_tasks(
        self,
        num_tasks_per_type: int = 10,
        methods: List = None,
        abox=False,
    ) -> TaskCollection:
        if methods is None:
            if abox:
                methods = [
                    self.extract_abox_task,
                ]
            else:
                methods = [
                    self.extract_taxon_constraint_task,
                    self.extract_indirect_superclasses_task,
                    self.extract_transitive_superclasses_task,
                    self.extract_most_recent_common_subsumers_task,
                    self.extract_subclass_of_expression_task,
                    self.extract_incoherent_ontology_task,
                ]
        objs = []
        for method in methods:
            for _n in range(num_tasks_per_type):
                task = method(select_random=True)
                task.obfuscated = self.obfuscate
                objs.append(task)
                logger.info(f"  {task.name}")
        tc = TaskCollection(tasks=objs)
        if self.obfuscated_curie_map:
            tc.obfuscated_curie_map = self.obfuscated_curie_map
        tc.name_map = self.name_map
        return tc

    def extract_ontology(
        self,
        terms: List[CURIE],
        roots: Optional[List[CURIE]] = None,
        predicates: Optional[List[PRED_CURIE]] = None,
        include_abox=False,
    ) -> Ontology:
        """
        Extract an ontology module following specified terms up the hierarchy.

        :param terms: leaf nodes to traverse up from
        :param roots: optional root nodes to stop at
        :param predicates: defaults to IS_A
        :return:
        """
        if predicates is None and not include_abox:
            predicates = [IS_A]
        adapter = self.adapter
        onts = list(adapter.ontologies())
        ancs = list(
            adapter.ancestors(terms, predicates=predicates, method=GraphTraversalMethod.HOP)
        )
        if NEVER_IN_TAXON in predicates:
            tax_ancs = list(adapter.ancestors(ancs, predicates=[IS_A]))
            ancs.extend(tax_ancs)
        if roots:
            roots_set = set(roots)
            ancs = [
                t
                for t in ancs
                if roots_set.intersection(adapter.ancestors(t, predicates=predicates))
            ]
        axioms = []
        already_have = set()
        terms_set = set(terms)
        used_predicates = set()
        if not ancs:
            raise ValueError(f"No ancestors found for {terms} over {predicates}")
        for t in ancs:
            for rel in adapter.relationships([t], predicates=predicates, include_abox=include_abox):
                if rel in already_have:
                    continue
                s, p, o = rel
                terms_set.add(s)
                terms_set.add(o)
                used_predicates.add(p)
                if roots and not set(adapter.ancestors(o, predicates=predicates)).intersection(
                    roots
                ):
                    continue
                axioms.append(self._axiom(rel, tbox=not include_abox))
                already_have.add(rel)
        if not axioms:
            raise ValueError(
                f"No axioms found for ancestors {ancs} over {predicates} (roots={roots})"
            )

        ontology = Ontology(
            name="-".join(onts), axioms=axioms, terms=list(terms_set), predicates=used_predicates
        )
        if include_abox:
            ontology.axioms.extend(list(self.extract_rbox()))
        if self.additional_axioms:
            ontology.axioms.extend(self.additional_axioms)
        return ontology

    def extract_indirect_superclasses_task(
        self,
        subclass: Optional[CURIE] = None,
        siblings: Optional[List[CURIE]] = None,
        roots: Optional[List[CURIE]] = None,
        predicates: Optional[List[PRED_CURIE]] = None,
        select_random=False,
        **kwargs,
    ) -> EntailedIndirectSuperClassTask:
        """
        Extract a task for finding all indirect superclasses of a class.

        >>> from oaklib import get_adapter
        >>> from ontogpt.ontex.extractor import OntologyExtractor
        >>> adapter = get_adapter("sqlite:obo:go")
        >>> extractor = OntologyExtractor(adapter=adapter)
        >>> task = extractor.extract_indirect_superclasses_task(
        ...    subclass="GO:0005634", siblings=["GO:0005773"], roots=["GO:0043226"]
        ... )

        :param subclass: the main focus of the query
        :param siblings: other terms to include (to make the task harder)
        :param roots: only include descendants of these terms
        :return: An EntailedIndirectSuperClassTask
        """
        if predicates is None:
            predicates = [IS_A]
        adapter = self.adapter
        if select_random:
            all_classes = list(adapter.entities(filter_obsoletes=True, owl_type=OWL_CLASS))
            subclass = random.choice(all_classes)
            siblings = random.sample(all_classes, 3)
        subclass_ancestors = list(adapter.ancestors(subclass, predicates=predicates))
        terms = [subclass] + siblings
        ontology = self.extract_ontology(terms, roots)
        if roots is not None:
            roots_set = set(roots)
        subclass_parents = {r[2] for r in adapter.relationships([subclass], predicates=predicates)}

        def _filter(anc: CURIE) -> bool:
            if anc == subclass:
                return True
            if anc in subclass_parents:
                return True
            if roots is not None:
                if not roots_set.intersection(adapter.ancestors(anc, predicates=predicates)):
                    return True

        filtered_ancestors = [anc for anc in subclass_ancestors if not _filter(anc)]
        answers = self._answers_from_ancestors(subclass, filtered_ancestors, predicates=predicates)
        task = EntailedIndirectSuperClassTask(
            ontology=ontology,
            query=Query(parameters=[self._name(subclass)]),
            answers=answers,
            **kwargs,
        )
        task.populate()

        return task

    def _answers_from_ancestors(
        self, start: CURIE, ends: List[CURIE], predicates: List[PRED_CURIE]
    ) -> List[ClassAnswer]:
        graph = self.adapter.ancestor_graph([start], predicates=predicates)
        answer_map = defaultdict(list)
        for _s, end, path in shortest_paths(graph, [start], ends, directed=True):
            axioms = []
            for i in range(len(path) - 1):
                axioms.append(self._axiom((path[i], IS_A, path[i + 1])))
            answer_map[end].append(Explanation(axioms=axioms))
        return [ClassAnswer(text=self._name(end), explanations=answer_map[end]) for end in ends]

    def extract_transitive_superclasses_task(
        self,
        subclass: CURIE = None,
        siblings: List[CURIE] = None,
        roots: Optional[List[CURIE]] = None,
        predicates: Optional[List[PRED_CURIE]] = None,
        select_random=False,
        **kwargs,
    ) -> EntailedTransitiveSuperClassTask:
        """
        Extract a task for finding all transitive superclasses of a class.

        >>> from oaklib import get_adapter
        >>> from ontogpt.ontex.extractor import OntologyExtractor
        >>> adapter = get_adapter("sqlite:obo:go")
        >>> extractor = OntologyExtractor(adapter=adapter)
        >>> task = extractor.extract_transitive_superclasses_task(
        ...    subclass="GO:0005634", siblings=["GO:0005773"], roots=["GO:0043226"]
        ... )

        :param subclass: the main focus of the query
        :param siblings: other terms to include (to make the task harder)
        :param roots: only include descendants of these terms
        :return: An EntailedIndirectSuperClassTask
        """
        if predicates is None:
            predicates = [IS_A]
        adapter = self.adapter
        if select_random:
            all_classes = list(adapter.entities(filter_obsoletes=True, owl_type=OWL_CLASS))
            subclass = random.choice(all_classes)
            siblings = random.sample(all_classes, 3)
        subclass_ancestors = list(adapter.ancestors(subclass, predicates=predicates))
        terms = [subclass] + siblings
        ontology = self.extract_ontology(terms, roots)
        answers = []
        if roots is not None:
            roots = list(set(roots))

        def _filter(anc: CURIE) -> bool:
            if anc == subclass:
                return True
            if roots is not None:
                roots_set = set(roots)
                if not roots_set.intersection(adapter.ancestors(anc, predicates=predicates)):
                    return True

        filtered_ancestors = [anc for anc in subclass_ancestors if not _filter(anc)]
        answers = self._answers_from_ancestors(subclass, filtered_ancestors, predicates=predicates)
        task = EntailedTransitiveSuperClassTask(
            ontology=ontology,
            query=Query(parameters=[self._name(subclass)]),
            answers=answers,
            **kwargs,
        )
        task.populate()
        return task

    def extract_abox_task(
        self,
        subject: CURIE = None,
        siblings: List[CURIE] = None,
        predicate: PRED_CURIE = None,
        select_random=False,
        **kwargs,
    ) -> ABoxTask:
        """Extract a task for finding all entailed edges."""
        adapter = self.adapter
        # TODO: this is currently necessary to get all entailed abox relationships
        include_tbox = True
        if select_random:
            all_entities = list(
                adapter.entities(filter_obsoletes=True, owl_type=OWL_NAMED_INDIVIDUAL)
            )
            all_direct_rels = list(
                adapter.relationships(
                    include_abox=True, include_tbox=include_tbox, include_entailed=False
                )
            )
            all_rels = list(
                adapter.relationships(
                    include_abox=True, include_tbox=include_tbox, include_entailed=True
                )
            )
            all_rels = [r for r in all_rels if r[2] in all_entities]
            if self.query_predicates:
                all_rels = [r for r in all_rels if r[1] in self.query_predicates]
            # for r in all_rels:
            #     print(r)
            all_indirect_rels = [r for r in all_rels if r not in all_direct_rels]
            candidate_rel = random.choice(all_indirect_rels)
            subject = candidate_rel[0]
            predicate = candidate_rel[1]
            siblings = random.sample(all_entities, 3)
        predicates = [predicate]
        relationships = list(
            adapter.relationships(
                [subject],
                predicates=predicates,
                include_abox=True,
                include_tbox=include_tbox,
                include_entailed=True,
            )
        )
        terms = [subject] + siblings
        # note: we need to extract over all predicates
        ontology = self.extract_ontology(terms, predicates=None, include_abox=True)
        # exclude reflexive
        ancestors = [r[2] for r in relationships if r[2] != subject]
        answers = self._answers_from_ancestors(subject, ancestors, predicates=predicates)
        task = ABoxTask(
            ontology=ontology,
            query=Query(parameters=[self._name(subject), self._name(predicate)]),
            answers=answers,
            **kwargs,
        )
        task.populate()
        return task

    def extract_most_recent_common_subsumers_task(
        self,
        subclass1: CURIE = None,
        subclass2: CURIE = None,
        siblings: List[CURIE] = None,
        roots: Optional[List[CURIE]] = None,
        predicates: Optional[List[PRED_CURIE]] = None,
        select_random=False,
        **kwargs,
    ) -> MostRecentCommonSubsumerTask:
        """Extract a task for finding all MRCAs of a pair of classes."""
        if predicates is None:
            predicates = [IS_A]
        adapter = self.adapter
        if select_random:
            all_classes = list(adapter.entities(filter_obsoletes=True, owl_type=OWL_CLASS))
            subclass1 = random.choice(all_classes)
            subclass2 = random.choice(all_classes)
            siblings = random.sample(all_classes, 2)
        terms = [subclass1, subclass2] + siblings
        ontology = self.extract_ontology(terms, roots)
        answers = []
        if not isinstance(adapter, SemanticSimilarityInterface):
            raise ValueError("Adapter must implement SemanticSimilarityInterface")
        mrcas = list(
            adapter.most_recent_common_ancestors(subclass1, subclass2, predicates=predicates)
        )
        for mrca in mrcas:
            explanations = [
                Explanation(
                    axioms=[
                        self._axiom((mrca, IS_A, subclass1)),
                        self._axiom((mrca, IS_A, subclass2)),
                    ]
                )
            ]
            answers.append(ClassAnswer(text=self._name(mrca), explanations=explanations))
        task = MostRecentCommonSubsumerTask(
            ontology=ontology,
            query=Query(parameters=[self._name(subclass1), self._name(subclass2)]),
            answers=answers,
            **kwargs,
        )
        task.populate()
        return task

    def extract_subclass_of_expression_task(
        self,
        superclass: CURIE = None,
        predicate: PRED_CURIE = None,
        siblings: List[CURIE] = None,
        predicates: Optional[List[PRED_CURIE]] = None,
        select_random=False,
        **kwargs,
    ) -> EntailedSubClassOfExpressionTask:
        adapter = self.adapter
        if predicate is None:
            predicate = PART_OF
        if not predicates:
            predicates = [IS_A, predicate]
        if select_random:
            all_classes = list(adapter.entities(filter_obsoletes=True, owl_type=OWL_CLASS))
            siblings = random.sample(all_classes, 2)
            n = 0
            while True:
                superclass = random.choice(all_classes)
                descendants = list(adapter.descendants(superclass, predicates=predicates))
                isa_descendants = list(adapter.descendants(superclass, predicates=[IS_A]))
                if (
                    len(descendants) < 15
                    and len(descendants) > 0
                    and len(descendants) != len(isa_descendants)
                ):
                    break
                n += 1
                if n > 100:
                    raise ValueError(
                        f"Could not find suitable parent (ontology MUST have {predicate}"
                    )
        logger.info(f"Extracting subclass of expression task for {superclass}, preds={predicates}")
        descendants = list(adapter.descendants(superclass, predicates=predicates))
        isa_descendants = list(adapter.descendants(superclass, predicates=[IS_A]))
        terms = descendants + siblings
        roots = [superclass] + siblings
        ontology = self.extract_ontology(terms, roots, predicates=predicates)
        answers = []
        if roots is not None:
            roots = list(set(roots))
        for desc in descendants:
            if desc == superclass:
                continue
            if desc in isa_descendants:
                # TODO: Reflexive scenario
                continue
            # if desc not in ontology.terms:
            #    continue
            explanations: List[str] = []
            answers.append(ClassAnswer(text=self._name(desc), explanations=explanations))
        task = EntailedSubClassOfExpressionTask(
            ontology=ontology,
            query=Query(parameters=[self._name(predicate), self._name(superclass)]),
            answers=answers,
            **kwargs,
        )
        task.populate()
        return task

    def extract_incoherent_ontology_task(
        self,
        incoherents: List[CURIE] = None,
        siblings: List[CURIE] = None,
        disjoints: List[Tuple[CURIE, CURIE]] = None,
        spiked_relationships: List[RELATIONSHIP] = None,
        roots: Optional[List[CURIE]] = None,
        select_random=False,
        **kwargs,
    ) -> OntologyCoherencyTask:
        """
        Extract task for testing ability to find incoherencies based on disjointness axioms.

        :param incoherents: expected incoherent classes
        :param siblings:
        :param disjoints: pairs of disjoint classes
        :param spiked_relationships: edges to insert into ontology
        :param roots:
        :return:
        """
        adapter = self.adapter
        if select_random:
            all_classes = list(adapter.entities(filter_obsoletes=True, owl_type=OWL_CLASS))
            siblings = random.sample(all_classes, 2)
            candidates = []
            for c in all_classes:
                parents = {rel[2] for rel in adapter.relationships(subjects=[c], predicates=[IS_A])}
                if len(parents) > 1:
                    candidates.append((c, parents))
            if len(candidates) == 0:
                raise ValueError("No suitable candidates")
            root_incoherent, parents = random.choice(candidates)
            incoherents = [
                random.choice(list(adapter.descendants(root_incoherent, predicates=[IS_A])))
            ]
            parents_list = list(parents)
            random.shuffle(parents_list)
            disjoints = [(parents_list[0], parents_list[1])]
        if not incoherents or not siblings or not disjoints:
            raise ValueError("Must specify incoherents, siblings, and disjoints")
        if not spiked_relationships:
            spiked_relationships = []
        terms = incoherents + siblings
        for s, _p, o in spiked_relationships:
            terms += [s, o]
        terms = list(set(terms))
        ontology = self.extract_ontology(terms, roots)
        for s, p, o in spiked_relationships:
            ontology.axioms.append(self._axiom((s, p, o)))
        for s, o in disjoints:
            ontology.axioms.append(self._axiom((s, DISJOINT_WITH, o)))
        # for s, p, o in spiked_relationships:
        #     if not isinstance(adapter, PatcherInterface):
        #         raise ValueError("Cannot patch ontology")
        #     adapter.apply_patch(kgcl.EdgeCreation(id='tmp', subject=s, predicate=p, object=o))
        answers = []
        if roots is not None:
            roots = list(set(roots))
        for incoherent in incoherents:
            ancestors = list(
                adapter.ancestors(incoherent, predicates=[IS_A], method=GraphTraversalMethod.HOP)
            )
            explanations = []
            # TODO:
            for s, o in disjoints:
                if s in ancestors and o in ancestors:
                    explanation = Explanation(
                        axioms=[
                            self._axiom((incoherent, IS_A, s)),
                            self._axiom((incoherent, IS_A, o)),
                        ]
                    )
                    explanations.append(explanation)
            if not explanations:
                explanations = [
                    Explanation(axioms=[self._axiom(rel) for rel in spiked_relationships])
                ]
            answers.append(ClassAnswer(text=self._name(incoherent), explanations=explanations))
        task = OntologyCoherencyTask(
            ontology=ontology,
            query=Query(),
            answers=answers,
            **kwargs,
        )
        task.populate()
        return task

    def extract_taxon_constraint_task(
        self,
        term: CURIE = None,
        taxon: CURIE = None,
        siblings: List[CURIE] = None,
        never_in=False,
        select_random=False,
        **kwargs,
    ) -> TaxonConstraintTask:
        adapter = self.adapter
        taxa = list(adapter.descendants("NCBITaxon:1", predicates=[IS_A]))
        all_terms = [
            t for t in adapter.entities(filter_obsoletes=True, owl_type=OWL_CLASS) if t not in taxa
        ]
        true_taxa = [t for t in taxa if "Union" not in t]
        relationships = list(adapter.relationships(predicates=TAXON_PREDICATES))
        never_in = [rel for rel in relationships if rel[1] == NEVER_IN_TAXON]
        if not never_in:
            raise ValueError("No never in taxon relationships")
        if select_random:
            if never_in or random.choice([True, False]):
                anc_term, p, direct_taxon = random.choice(never_in)
            else:
                anc_term, p, direct_taxon = random.choice(
                    [r for r in relationships if r[1] != NEVER_IN_TAXON]
                )
            candidates = list(adapter.descendants(anc_term))
            candidate_taxa = list(adapter.descendants(direct_taxon))
            candidate_taxa = [t for t in candidate_taxa if t in true_taxa]
            term = random.choice(candidates)
            taxon = random.choice(candidate_taxa)
            remaining = list(set(candidates) - set(all_terms))
            siblings = random.sample(remaining, min(2, len(remaining)))
        if not term or not taxon:
            raise ValueError("Must specify term and taxon")
        terms = true_taxa + [term] + siblings
        ontology = self.extract_ontology(terms, predicates=TAXON_PREDICATES + [IS_A, PART_OF])
        for t in true_taxa:
            children = {rel[0] for rel in adapter.relationships(objects=[t], predicates=[IS_A])}
            children_list = [c for c in children if c in true_taxa]
            if len(children_list) > 1:
                for i in range(len(children_list)):
                    for j in range(i + 1, len(children_list)):
                        ontology.axioms.append(
                            self._axiom((children_list[i], DISJOINT_WITH, children_list[j]))
                        )
        if not isinstance(adapter, TaxonConstraintInterface):
            raise ValueError("Cannot evaluate taxon constraints")
        st = SubjectTerm(term, label=adapter.label(term))
        st.present_in.append(TaxonConstraint(taxon=Taxon(taxon)))
        st = adapter.eval_candidate_taxon_constraint(st, predicates=[IS_A, PART_OF])
        if st.unsatisfiable:
            validity = "INVALID"
        else:
            validity = "VALID"
        answers = []
        answers.append(ClassAnswer(text=validity, explanations=[]))
        task = TaxonConstraintTask(
            ontology=ontology,
            query=Query(parameters=[self._name(term), self._name(taxon)]),
            answers=answers,
            **kwargs,
        )
        task.populate()
        return task

    def _axiom(self, rel: RELATIONSHIP, tbox=True) -> Axiom:
        s, p, o = rel
        s_n, p_n, o_n = self._name(s), self._name(p), self._name(o)
        if p == IS_A:
            return Axiom(text=f"{s_n} SubClassOf {o_n}")
        elif p == DISJOINT_WITH:
            return Axiom(text=f"{s_n} DisjointWith {o_n}")
        elif p == NEVER_IN_TAXON:
            return Axiom(text=f"{s_n} never_in_taxon {o_n}")
        elif p in [IN_TAXON, ONLY_IN_TAXON]:
            return Axiom(text=f"{s_n} only_in_taxon {o_n}")
        elif tbox:
            return Axiom(text=f"{s_n} SubClassOf {p_n} Some {o_n}")
        else:
            return Axiom(text=f"{s_n} {p_n} {o_n}")

    def _name(self, curie: CURIE) -> str:
        lbl = self.adapter.label(curie)
        if lbl is None:
            return curie
        else:
            lbl = re.sub(r"\W+", "_", lbl)
        lbl = inflection.camelize(lbl)
        self.name_map[curie] = lbl  # type: ignore
        if self.use_identifiers:
            return curie
        elif self.obfuscate:
            obfuscated = base64.b64encode(curie.encode("utf-8")).decode("utf-8")
            self.obfuscated_curie_map[obfuscated] = curie  # type: ignore
            return obfuscated
        else:
            return lbl

    def extract_rbox(self) -> Iterable[Axiom]:
        if not isinstance(self.adapter, SqlImplementation):
            raise ValueError("Only SQL adapters supported")
        session = self.adapter.session
        for name, curie in [
            ("TransitiveProperty", OWL_TRANSITIVE_PROPERTY),
            ("SymmetricProperty", OWL_SYMMETRIC_PROPERTY),
        ]:
            q = session.query(RdfTypeStatement).filter(RdfTypeStatement.object == curie)
            for row in q:
                yield Axiom(text=f"{self._name(row.subject)} type {name}")
        for row in session.query(Statements).filter(Statements.predicate == SUBPROPERTY_OF):
            yield Axiom(text=f"{self._name(row.subject)} SubPropertyOf {self._name(row.object)}")
        for row in session.query(Statements).filter(Statements.predicate == INVERSE_OF):
            yield Axiom(text=f"{self._name(row.subject)} SubPropertyOf {self._name(row.object)}")
        for row in session.query(Statements).filter(
            Statements.predicate == "owl:propertyChainAxiom"
        ):
            entailed_pred = row.subject
            bnode = row.object
            chain = []
            for inner_row in session.query(RdfListMemberStatement.object).filter(
                RdfListMemberStatement.subject == bnode
            ):
                chain.append(inner_row[0])
            expr = " o ".join(self._name(p) for p in chain)
            yield Axiom(text=f"{expr} SubPropertyOf {self._name(entailed_pred)}")



"""ontogpt package."""
from importlib import metadata
from pathlib import Path

from yaml import safe_load

# Load the model list and define the default model
rel_path = Path(__file__).resolve()
models_path = rel_path.parent / "models.yaml"

with open(models_path, "r") as models_file:
    MODELS = (safe_load(models_file))["models"]
for model in MODELS:
    if "is_default" in model:
        DEFAULT_MODEL_DETAILS = model
        DEFAULT_MODEL = model["alternative_names"][0]
        break

try:
    __version__ = metadata.version(__name__)
except metadata.PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"  # pragma: no cover



"""HuggingFace Hub Client."""
import logging
from dataclasses import dataclass

from langchain import HuggingFaceHub, LLMChain, PromptTemplate
from oaklib.utilities.apikey_manager import get_apikey_value

# Note: See https://huggingface.co/models?pipeline_tag=text-generation&sort=downloads
# for all relevant models


@dataclass
class HFHubClient:
    """A client for the HuggingFace Hub API."""

    try:
        api_key = get_apikey_value("hfhub-key")
    except ValueError:
        logging.error("HuggingFace Hub API key not found. See README.")

    def get_model(self, modelname: str) -> HuggingFaceHub:
        """Retreive a model from the Hub, given its repository name.

        Returns a model object of type
        langchain.llms.huggingface_hub.HuggingFaceHub
        """
        model = HuggingFaceHub(
            repo_id=modelname,
            verbose=True,
            model_kwargs={"temperature": 0.2, "max_length": 500},
            huggingfacehub_api_token=self.api_key,
            task="text-generation",
        )

        return model

    def query_hf_model(self, llm, prompt_text):
        """Interact with a GPT4All model."""
        logging.info(f"Complete: prompt[{len(prompt_text)}]={prompt_text[0:100]}...")

        template = """{prompt_text}"""

        prompt = PromptTemplate(template=template, input_variables=["prompt_text"])

        llm_chain = LLMChain(prompt=prompt, llm=llm)

        try:
            raw_output = llm_chain.run({"prompt_text": prompt_text})
        except ValueError as e:
            logging.error(e)
            raw_output = ""

        return raw_output



"""Resolver engine."""
from typing import Optional, Type, Union

from class_resolver import ClassResolver

from ontogpt.engines.enrichment import EnrichmentEngine
from ontogpt.engines.halo_engine import HALOEngine  # type: ignore
from ontogpt.engines.knowledge_engine import KnowledgeEngine
from ontogpt.engines.spires_engine import SPIRESEngine

resolver = ClassResolver([SPIRESEngine, HALOEngine, EnrichmentEngine], base=KnowledgeEngine)


def create_engine(
    template: str, engine: Optional[Union[str, Type]] = None, **kwargs
) -> Union[KnowledgeEngine, SPIRESEngine]:
    """Create a knowledge engine."""
    if engine is None:
        engine = SPIRESEngine
    if isinstance(engine, str):
        engine = resolver.get_class(engine)(**kwargs)
    if engine is not None and not isinstance(engine, str):
        return engine(template, **kwargs)
    else:
        return SPIRESEngine  # type: ignore



"""Markdown exporter."""
from dataclasses import dataclass
from io import BytesIO, StringIO, TextIOWrapper
from pathlib import Path
from typing import Optional, TextIO, Union

import pydantic
import yaml
from linkml_runtime import SchemaView

from ontogpt.io.exporter import Exporter, is_curie
from ontogpt.templates.core import ExtractionResult


@dataclass
class MarkdownExporter(Exporter):
    def export(
        self,
        extraction_output: ExtractionResult,
        output: Union[str, Path, TextIO, BytesIO],
        schemaview: Optional[SchemaView] = None,
    ):  # type: ignore
        if isinstance(output, Path):
            output = open(str(output), "w", encoding="utf-8")
        if isinstance(output, str):
            output = StringIO(output)
        if isinstance(output, BytesIO):
            output = TextIOWrapper(output, encoding="utf-8")
        output.write(f"# {extraction_output.input_id}\n\n")
        output.write("## Input\n\n")
        for block in extraction_output.input_text.split("\n"):
            output.write(f"_{block.replace('_', '')}_\n\n")
        output.write("## Results\n\n")
        obj = extraction_output.extracted_object
        self.export_object(obj, extraction_output, output, -1)
        output.write("\n\nYAML:\n\n")
        self.details(yaml.dump(extraction_output.dict()), output, code="yaml")
        output.write("\n\nPrompt:\n\n")
        self.details(extraction_output.prompt, output)
        output.write("\n\nCompletion:\n\n")
        self.details(extraction_output.raw_completion_output, output)

    def export_object(
        self,
        obj: pydantic.BaseModel,
        extraction_output: ExtractionResult,
        output: TextIO,
        indent: int,
    ):
        for field in obj.__fields__.values():
            if indent < 0:
                output.write(f"\n\n### {field.name}\n\n")
            else:
                output.write(f"\n{'  ' * indent}- {field.name}:")
            value = getattr(obj, field.name)
            if isinstance(value, pydantic.BaseModel):
                self.export_object(value, extraction_output, output, indent + 1)
            elif isinstance(value, list):
                for item in value:
                    if isinstance(item, pydantic.BaseModel):
                        self.export_object(
                            item,
                            extraction_output=extraction_output,
                            output=output,
                            indent=indent + 1,
                        )
                    else:
                        self.export_atom(item, extraction_output, output, indent + 1)
            else:
                self.export_atom(value, extraction_output, output, indent + 1)
        output.write("\n")

    def export_atom(self, value, extraction_output: ExtractionResult, output: TextIO, indent: int):
        matches = [
            ne for ne in extraction_output.named_entities if ne.id == value and is_curie(ne.id)
        ]
        output.write(f"\n{'  ' * indent}- ")
        if matches:
            match = matches[0]
            output.write(f"{match.label} {self.link(match.id)}")
        else:
            output.write(f"{value}")
        output.write("\n")

    def details(self, text: str, output: TextIO, code: str = ""):
        output.write("<details>\n")
        output.write(f"```{code}\n")
        output.write(text)
        output.write("\n```\n")
        output.write("\n</details>\n")

    def link(self, curie: str) -> str:
        return f"[{curie}](https://bioregistry.io/{curie})"



from __future__ import annotations
from datetime import datetime, date
from enum import Enum
from typing import List, Dict, Optional, Any, Union
from pydantic import BaseModel as BaseModel, Field
import sys
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


metamodel_version = "None"
version = "None"

class ConfiguredBaseModel(BaseModel,
                validate_assignment = True,
                validate_default = True,
                extra = 'forbid',
                arbitrary_types_allowed = True,
                use_enum_values = True):
    pass


class NullDataOptions(str, Enum):
    
    
    UNSPECIFIED_METHOD_OF_ADMINISTRATION = "UNSPECIFIED_METHOD_OF_ADMINISTRATION"
    
    NOT_APPLICABLE = "NOT_APPLICABLE"
    
    NOT_MENTIONED = "NOT_MENTIONED"
    
    

class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    input_id: Optional[str] = Field(None)
    input_title: Optional[str] = Field(None)
    input_text: Optional[str] = Field(None)
    raw_completion_output: Optional[str] = Field(None)
    prompt: Optional[str] = Field(None)
    extracted_object: Optional[Any] = Field(None, description="""The complex objects extracted from the text""")
    named_entities: Optional[List[Any]] = Field(default_factory=list, description="""Named entities extracted from the text""")
    

class NamedEntity(ConfiguredBaseModel):
    
    id: Optional[str] = Field(None, description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class DiagnosticProcedure(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Phenotype(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class ClinicalAttribute(NamedEntity):
    
    unit: Optional[str] = Field(None, description="""the unit used to measure the attribute""")
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Quality(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Unit(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class CompoundExpression(ConfiguredBaseModel):
    
    None
    

class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""")
    

class DiagnosticProceduretoPhenotypeAssociation(Triple):
    """
    A triple representing a relationship between a diagnostic procedure and an associated phenotype, e.g., \"blood pressure measurement\" is associated with \"high blood pressure\".
    """
    subject: Optional[str] = Field(None, description="""A diagnostic procedure yielding a result, which in turn may be interpreted as a phenotype. Procedures include \"heart rate measurement\", \"blood pressure measurement\", \"oxygen saturation measurement\", etc. In practice, procedures may be named based on what they measure, with the \"measurement\" part left implicit.""")
    predicate: Optional[str] = Field(None, description="""The relationship type, e.g. RELATED_TO""")
    object: Optional[List[str]] = Field(default_factory=list, description="""The observable physical or biochemical characteristics of a patient. Not equivalent to a disease state, but may contribute to a diagnosis.""")
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the procedure.""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the phenotype.""")
    

class DiagnosticProceduretoAttributeAssociation(Triple):
    """
    A triple representing a relationship between a diagnostic procedure and a measured attribute, e.g., \"blood pressure measurement\" is associated with \"blood pressure\" (or in OBA, something like OBA:VT0000183, \"blood pressure trait\").
    """
    subject: Optional[str] = Field(None, description="""A diagnostic procedure yielding a result, which in turn may be interpreted as a phenotype. Procedures include \"heart rate measurement\", \"blood pressure measurement\", \"oxygen saturation measurement\", etc. In practice, procedures may be named based on what they measure, with the \"measurement\" part left implicit.""")
    predicate: Optional[str] = Field(None, description="""The relationship type, e.g. RELATED_TO""")
    object: Optional[List[str]] = Field(default_factory=list, description="""Any measurable clinical attribute.""")
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the procedure.""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the phenotype.""")
    

class TextWithTriples(ConfiguredBaseModel):
    
    publication: Optional[Publication] = Field(None)
    triples: Optional[List[Triple]] = Field(default_factory=list)
    

class RelationshipType(NamedEntity):
    
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class ProcedureToPhenotypePredicate(RelationshipType):
    """
    A predicate for procedure to phenotype relationships, defining \"this procedure is intended to provide support for/against this phenotype\".
    """
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class ProcedureToAttributePredicate(RelationshipType):
    """
    A predicate for procedure to attribute relationships, defining \"this procedure is a measurement of this attribute\".
    """
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")
    

class Publication(ConfiguredBaseModel):
    
    id: Optional[str] = Field(None, description="""The publication identifier""")
    title: Optional[str] = Field(None, description="""The title of the publication""")
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""")
    combined_text: Optional[str] = Field(None)
    full_text: Optional[str] = Field(None, description="""The full text of the publication""")
    

class AnnotatorResult(ConfiguredBaseModel):
    
    subject_text: Optional[str] = Field(None)
    object_id: Optional[str] = Field(None)
    object_text: Optional[str] = Field(None)
    


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
ExtractionResult.model_rebuild()
NamedEntity.model_rebuild()
DiagnosticProcedure.model_rebuild()
Phenotype.model_rebuild()
ClinicalAttribute.model_rebuild()
Quality.model_rebuild()
Unit.model_rebuild()
CompoundExpression.model_rebuild()
Triple.model_rebuild()
DiagnosticProceduretoPhenotypeAssociation.model_rebuild()
DiagnosticProceduretoAttributeAssociation.model_rebuild()
TextWithTriples.model_rebuild()
RelationshipType.model_rebuild()
ProcedureToPhenotypePredicate.model_rebuild()
ProcedureToAttributePredicate.model_rebuild()
Publication.model_rebuild()
AnnotatorResult.model_rebuild()
    
